import os
import sys
import copy
import shutil
import pickle
import random
import subprocess
import numpy as np
import multiprocessing as mp
from tqdm import tqdm
from graph_tool.all import *
from datetime import datetime
from itertools import combinations, chain
from scipy.sparse import lil_matrix, csr_matrix, save_npz, load_npz, triu
from collections import defaultdict, Counter
from mycotools.lib.biotools import gff2list, dict2fa
from mycotools.lib.kontools import write_json, read_json, collect_files, \
                                   checkdir, eprint
from orthocluster.orthocluster.lib import input_parsing, treecalcs, evo_conco, output_data


def hash_hgx(gff_path, ome, hgx_genes, gene2hg, clusplusminus):
    """sliding window of size clusplusminus, identify sequences that may have
    hgs that may be significant, grab their window, and check for a significant
    higher order og combo in the locus"""
        
    gff_list, gene_dict = gff2list(gff_path), defaultdict(list)
    cds_dict = input_parsing.compileCDS(gff_list, os.path.basename(gff_path).replace('.gff3',''))
                
    hgx_dict = {}   
    for scaf in cds_dict: 
        for i0, seq0 in enumerate(cds_dict[scaf]):
            if seq0 in hgx_genes: # if the og is part of a significant seed
            # locus 
                if i0 < clusplusminus: # is i0 - clusplusminus < 0 ?
                    locus = cds_dict[scaf][:i0+clusplusminus+1] # then gather
                    # all the beginning
                else:
                    locus = cds_dict[scaf][i0-clusplusminus:i0+clusplusminus+1]
                    # instead get the +/- and adjust for python
                og0 = gene2hg[seq0]
                for hgx in hgx_genes[seq0]:
                    if hgx not in hgx_dict:
                        hgx_dict[hgx] = {og: [] for og in hgx}
                    try:
                        hgx_dict[hgx][og0].append(seq0) # add the sequence to the hgx_dict
                    except KeyError:
                        pass
                    start, end = None, None
                    for i1, seq1 in enumerate(locus): # for each index and sequence
                    # in the locus
                        try:
                            og1 = gene2hg[seq1]
                        except KeyError: # missing entry
                            continue
                        if og1 in set(hgx) and start is None: # if the og is
                        # in the sig clus and we haven't started
                            start = i1 # this is the start index
                            hgx_dict[hgx][og1].append(seq1)

                        elif og1 in set(hgx): # otherwise if it is in the
                        # sig clus label this the other border unless another is 
                        # found
                            end = i1 + 1
                            hgx_dict[hgx][og1].append(seq1)
                    for gene in locus[start:end]:
                        gene_dict[gene].append(hgx)

    gene_tup = tuple([
        tuple([gene, tuple(sorted(vals))]) \
        for gene, vals in gene_dict.items()
        ])
    # ((gene, {gene: [HGx...]}))

    hgx_tup = tuple([
        tuple([
            hgx,
            tuple([tuple([og, tuple(sorted(set(seqs)))]) \
                for og, seqs in hgs.items()]) \
            ])
            for hgx, hgs in hgx_dict.items()
        ])

    return ome, gene_tup, hgx_tup


def find_hgx_pairs(gene2hgx, ome, hgx2i, pairsDict, min_hgx_overlap = 2):
    # is this going to merge hgxs that overlap by one gene in min_hgx_overlap loci?
    hgx_pairs_raw = []
    # check for gene overlap in hgxs
    for hgxs in list(gene2hgx.values()):
        # add each possible hgxpair
        hgx_pairs_raw.extend(list(combinations(sorted(hgxs), 2)))

    hgx_pairs_count = [tuple((hgx2i[y] for y in x)) \
                       for x, g in Counter(hgx_pairs_raw).items() \
                       if g >= min_hgx_overlap]
    for hgxpair in hgx_pairs_count:
        pairsDict[hgxpair].append(ome)
    return pairsDict


def read_mci_rows(rows_file):
    mci2pre = {}
    with open(rows_file, 'r') as raw:
        for line in raw:
            mci, pre = [int(x) for x in line.rstrip().split()]
            mci2pre[mci] = pre
    return mci2pre


def MCL(adj_path, clus_file, inflation = 1.5, threads = 1):

    mci_file = adj_path[:-4] + '.mci'
    rows_file = os.path.dirname(adj_path) + '/mcl_rows.tsv'
    if not os.path.isfile(mci_file):
        subprocess.call(['mcxload', '-abc', adj_path, '-o',
                         mci_file + '.tmp', '--write-binary',
                         '-write-tab', rows_file, '--stream-mirror'],
                         stdout = subprocess.DEVNULL, 
                         stderr = subprocess.DEVNULL)
        shutil.move(mci_file + '.tmp', mci_file)
    subprocess.call([
        'mcl', mci_file, '-I', str(inflation),
        '-o', clus_file + '.tmp', '-te', str(threads)
        ], stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL
        )
    subprocess.call(['mcxdump', '-icl', clus_file + '.tmp',
                     '-o', clus_file, '--dump-pairs'],
                    stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL)
    os.remove(clus_file + '.tmp')


def sorensen(interlen, len0, len1):
    return 2*interlen/(len0 + len1)
def jaccard(interlen, len0, len1):
    return interlen/(len0 + len1 - interlen)
def overlap(interlen, len0, len1):
    return interlen/min([len0, len1])

def calc_loc2loc_sim_mat(
    i0, i1, index, loc0, loc1, hgL0, hgL1, coef, blast_ids,
    min_hlg_id = 0.15
    ):

    # Identify the overlapping HGs
    hg_dict = defaultdict(lambda: [[], []])
    for i, hg in enumerate(hgL0):
        hg_dict[hg][0].append(loc0[i])
    for i, hg in enumerate(hgL1):
        hg_dict[hg][1].append(loc1[i])
    hg_dict = {k: v for k, v in hg_dict.items() \
               if v[0] and v[1]}

    scores = []
    for hg, gene_lists in hg_dict.items():
        scores.append([])
        # for each pairwise gene comparison
        for gene0 in gene_lists[0]:
            for gene1 in gene_lists[1]:
                try:
                    scores[-1].append(blast_ids[hg][gene0][gene1])
                # if the genes are missing its a 0
                # the alignment algorithm MUST output enough alignments
                # for this is to reliably work on widely dispersed clusters
                except KeyError: 
                    pass
    scores = [x for x in scores if x]
    if scores:
        maxScores = [max(i) for i in scores]
        hlg_id = (sum(maxScores)/len(maxScores)) * coef
        if hlg_id > min_hlg_id:
            return f'{i0 + index}\t{i1 + index}\t{hlg_id}\n'
        else:
            return ''



def acquire_clus_hlg_sim_noq(
    i0, i1, index, loc0, loc1, hgL0, hgL1, set0, set1, blast_ids,
    simfun = overlap, min_hlg_id = 0.15
    ):

    # Identify the overlapping HGs
    intersection = set0.intersection(set1)
    inter_list = list(intersection)
    hg_dict = {k: [[], []] for k in inter_list}

    # populate an hg2gene dict for ome0 and ome1
    for i, hg in enumerate(hgL0):
        if hg in intersection:
            hg_dict[hg][0].append(loc0[i])
    for i, hg in enumerate(hgL1):
        if hg in intersection:
            hg_dict[hg][1].append(loc1[i])

    scores = []
    for hg in inter_list:
        # check to verify the genes are recovered
        scores.append([])
        # for each pairwise gene comparison
        for gene0 in hg_dict[hg][0]:
            for gene1 in hg_dict[hg][1]:
                try:
                    scores[-1].append(blast_ids[hg][gene0][gene1])
                # if the genes are missing its a 0
                # the alignment algorithm MUST output enough alignments
                # for this is to reliably work on widely dispersed clusters
                except KeyError: 
                    scores[-1].append(0)
    scores = [x for x in scores if x]
    if scores:
        maxScores = [max(i) for i in scores]
        coef = simfun(len(scores), len(set0), len(set1))
        hlg_id = (sum(maxScores)/len(maxScores)) * coef
        if hlg_id > min_hlg_id:
            return f'{i0 + index}\t{i1 + index}\t{hlg_id}\n'
        else:
            return ''


def rnd2_gen_blastids_mp(hg, group_dict, hgx_dir, grp_dir, minid):#, blast_ids_mngr):
    gene_set = set(chain(*list(group_dict.values())))
    gene_len = len(gene_set)
    algn_base = f'{hgx_dir}{hg}.out'
    json_out = f'{grp_dir}/algn/{hg}.json'
    if not os.path.isfile(json_out):
        try:
            hg_dict = parse_blast(algn_base, gene_len, gene_set, minid)
            new_hg_dict = defaultdict(dict)
            for g, genes in group_dict.items():
                for g0, g1 in combinations(sorted(genes), 2):
                    try:
                        new_hg_dict[g0][g1] = hg_dict[g0][g1]
                    except KeyError:
                        pass
                    try:
                        new_hg_dict[g1][g0] = hg_dict[g1][g0] 
                    except KeyError:
                        pass
            write_json(new_hg_dict, json_out + '.tmp')
            os.rename(json_out + '.tmp', json_out)
        except FileNotFoundError:
            pass



def gen_blastids_mp(hg, genes, hgx_dir, grp_dir, minid):#, blast_ids_mngr):
    # CURRENTLY NOT SETUP FOR ANY OTHER CLANS THAN 0
    # would need to incorporate clanI in the output json
    # or else it runs the risk of using a failed runs' old jsons
    # cannot multiprocess the runs either
    gene_set = set(genes)
    gene_len = len(gene_set)
    algn_base = f'{hgx_dir}{hg}.out'
    json_out = f'{grp_dir}/algn/{hg}.json'
    if not os.path.isfile(json_out):
        try:
            hg_dict = parse_blast(algn_base, gene_len, gene_set, minid)
            if hg_dict is not None:
                write_json(hg_dict, json_out + '.tmp')
                os.rename(json_out + '.tmp', json_out)
    #        blast_ids_mngr[hg].update(hg_dict)
        except FileNotFoundError:
            pass
 #       print(f'\t\t\t\t\tAttempting to BLAST {hg}', flush = True)
#        blast = run_hgx_blast(hg_dir, hg, genes, hgx_dir, 
  #                    blastp = 'blastp', cpus = 1)
   #     if blast:
    #        blast_ids_mngr[hg].update(parse_blast(algn_base, 
     #                                             gene_len, gene_set))


def read_hlg_sim_jsons(hlg_dir):
    jsons = collect_files(hlg_dir, 'json')
    blast_ids = {}
    for j_f in jsons:
        hg = os.path.basename(j_f).replace('.json','')
        blast_ids[int(hg)] = read_json(j_f)
    return blast_ids


def parse_blast(algn_base, gene_len, gene_set, minid = 30):
    hg_dict = defaultdict(dict)
    with open(algn_base, 'r') as raw:
        for line in raw:
            d = line.rstrip().split()
            # adjust to just reference d when safe
            try:
                q, s, pident = d[0], d[1], d[-1]
            except IndexError:
                eprint(f'ERROR: malformatted alignment: {os.path.basename(algn_base)}', 
                       flush = True) 
                return False
            if {q, s}.issubset(gene_set):
                # NEED a no self hits option in mmseqs or blast to avoid
                if q != s:
                    pid = float(pident)
                    if pid >= minid:
                        hg_dict[q][s] = pid/100 # adjust diamond to decimal
            if len(hg_dict) == gene_len:
                break
    return hg_dict


def hlg_sim_write_mngr(Q, sim_file):
    x = Q.get()
    with open(sim_file, 'a') as out:
        while x:
            out.write(x)
            x = Q.get()
            out.flush()


def calc_sim_by_row(adj_mat, hgLoci, loci, blast_ids, 
                    index, Q, max_complete, min_hlg_id, simfun):
 #   adj_mat = np.load(f'{mat_dir}{rowI}.npz')
    for ti0, overlap_loci in enumerate(adj_mat):
        i0 = ti0 + max_complete
        loc0, hgL0 = loci[i0], hgLoci[i0]
        sHGl0 = set([x for x in hgL0 if x is not None])
        for ti1 in overlap_loci:
            if ti1 > ti0:
                i1 = ti1 + max_complete
                loc1, hgL1 = loci[i1], hgLoci[i1]
                sHGl1 = set([x for x in hgL1 if x is not None])
                data = acquire_clus_hlg_sim_noq(i0, i1, index, loc0,
                                                loc1, hgL0, hgL1,
                                                sHGl0, sHGl1, blast_ids,
                                                min_hlg_id = min_hlg_id,
                                                simfun = simfun)
                if data:
                    Q.put(data)



def hlg_sim_mp_mngr(adj_mat, sim_file, max_complete, loci, hgLoci, index, 
                    blast_ids, hlg_dir, mat_dir, min_hlg_id, simfun, 
                    chunkscale = 100, cpus = 1):
    if not os.path.isdir(mat_dir):
        os.mkdir(mat_dir)
    
    # break up the adj_mat by chunksize and offload to file while preparing cmds
    row_len = len(adj_mat)
    chunksize = round(row_len/chunkscale)
    cmds = []
    Q = mp.Manager().Queue()
    for i, row in enumerate(range(0, row_len, chunksize)):
        if len(adj_mat) >= row + chunksize:
            sub_mat = [list(adj_mat[x]) for x in range(row, row + chunksize)]
        else:
            sub_mat = [list(adj_mat[x]) for x in range(row, len(adj_mat))]
#        np.save(f'{mat_dir}{i}.npz', sub_mat)
        locIs = [x for x in list(set(chain(*sub_mat))) if x >= row]
        t_hg_loc = {locI: hgLoci[locI + max_complete] for locI in locIs}
        t_loc = {locI: loci[locI + max_complete] for locI in locIs}
        t_hgs = set(chain(*list(t_hg_loc.values())))
        if None in t_hgs:
            t_hgs.remove(None)
        t_ids = {hg: blast_ids[hg] for hg in list(t_hgs) if hg in blast_ids}
        cmds.append((sub_mat, t_hg_loc, t_loc, t_ids, 
                     index, Q, max_complete, min_hlg_id, simfun))

    write_proc = mp.Process(target=hlg_sim_write_mngr, args=(Q, sim_file))
    write_proc.start()

    with mp.get_context('forkserver').Pool(processes = cpus - 1) as pool:
        pool.starmap(calc_sim_by_row, tqdm(cmds, total = len(cmds)))
        pool.close()
        pool.join()

    Q.put(None)
    write_proc.close()
    write_proc.join()

def calc_overlap_matrix(adj_arr, clan_arr):
    #INCORRECT! DOESN'T CURRENTLY USE MINIMUM!!!!!!
    return adj_arr/clan_arr.sum(axis=1).T

def calc_sorensen_matrix_sparse(adj_arr, clan_arr):
    return 2 * adj_arr/(clan_arr.sum(axis=1) + clan_arr.sum(axis=1).T)

def calc_sorensen_array(adj_arr, clan_arr):
    return 2 * adj_arr/(clan_arr.sum(axis=1) + np.array([clan_arr.sum(axis=1)]).T)

def matrix_sim_mngr(s_mat, out, loci, hgLoci, max_complete,
                    blast_ids, index, min_hlg_id):
    for ti0, row in enumerate(s_mat):
        i0 = ti0 + max_complete
        loc0, hgL0 = loci[i0], hgLoci[i0]
        for ti1, v in enumerate(row):
            i1 = ti1 + max_complete
            loc1, hgL1 = loci[i1], hgLoci[i1]
            data = calc_loc2loc_sim_mat(i0, i1, index, loc0,
                                        loc1, hgL0, hgL1,
                                        v, blast_ids,
                                        min_hlg_id = min_hlg_id)
            if data:
                out.write(data)
        out.flush()


def rnd2_loc2loc_sim_mngr(clan_arr, finished_file, max_complete, loci, hgLoci,
                     index, blast_ids, min_hlg_id, simfun):
    adj_arr = np.triu(clan_arr @ clan_arr.T)
    adj_arr[adj_arr < 2] = 0
    s_mat = calc_sorensen_array(adj_arr, clan_arr)
    np.fill_diagonal(s_mat, 0)
    with open(f'{finished_file}.w', 'a') as out:
        matrix_sim_mngr(s_mat, out, loci,
                        hgLoci, max_complete, blast_ids, index, min_hlg_id)



def loc2loc_sim_mngr(clan_arr, finished_file, max_complete, loci, hgLoci,
                     index, blast_ids, min_hlg_id, simfun):
    clan_arr = clan_arr.tocsr()
    # a matrix representing the cardinality of the intersection of hg_locs
    adj_arr = csr_matrix(triu(clan_arr @ clan_arr.transpose()))
    # less than two gene overlap is set to 0
    adj_arr.data[adj_arr.data < 2] = 0

    hits = np.split(adj_arr.indices, adj_arr.indptr)[1:-1]
    with open(f'{finished_file}.w', 'a') as out:
        for ti0, overlap_loci in enumerate(hits):
            i0 = ti0 + max_complete
            loc0, hgL0 = loci[i0], hgLoci[i0]
            sHGl0 = set([x for x in hgL0 if x is not None])
            for ti1 in overlap_loci[:]:
                i1 = ti1 + max_complete
                loc1, hgL1 = loci[i1], hgLoci[i1]
                sHGl1 = set([x for x in hgL1 if x is not None])
                data = acquire_clus_hlg_sim_noq(i0, i1, index, loc0,
                                                loc1, hgL0, hgL1,
                                                sHGl0, sHGl1, blast_ids,
                                                min_hlg_id = min_hlg_id,
                                                simfun = simfun)
                if data: 
                    out.write(data)
            out.flush()

def open_prev_res(finished_file, index):
    # identify the last definitively known clan
    with open(f'{finished_file}.w1', 'w') as out:
        with open(f'{finished_file}.w', 'r') as raw:
            data, max_complete = [], 0
            for line in raw:
                d = line.rstrip().split()
                try:
                    if d[0] != data[-1][0]:
                        out.write('\n'.join([' '.join(x) for x in data]) \
                                + '\n')
                        out.flush()
                        max_complete = int(data[0][0]) - index
                        data = []
                except IndexError: # first line
                    pass
                data.append(d)
    shutil.move(f'{finished_file}.w1', f'{finished_file}.w')
    return max_complete

def rnd2_loc2loc_mngr(hlg_dir, loci, hg_loci,
                     hgx_dir, groups, minid, min_hlg_id, simfun, cpus = 1):

    finished_file = f'{hlg_dir}loci.adj.tmp'
    # is this clan already completed?
    if os.path.isfile(finished_file) or os.path.isfile(finished_file + '.r'):
        return
    # was this clan partially written?
    elif os.path.isfile(f'{finished_file}.w'):
        os.remove(f'{finished_file}.w')

    # only examine loci that have not been completed
    ghg2i = defaultdict(dict)
    blast_hash = defaultdict(lambda: defaultdict(list))
    for g,v in enumerate(groups):
        i0, i1 = v
        for i2 in range(i0, i1):
            locus = loci[i2]
            hgs = hg_loci[i2]
            for i4, hg in enumerate(hgs):
                if hg is not None:
                    blast_hash[hg][g].append(locus[i4])
                    if hg not in ghg2i[g]:
                        ghg2i[g][hg] = len(ghg2i[g])

    clan_arrs = []
    for g, v in enumerate(groups):
        i0, i1 = v
        clan_arr = lil_matrix((i1 - i0, len(ghg2i[g])), dtype=bool)
#        clan_arr = np.zeros((i1-i0, len(ghg2i[g])), dtype=np.int8)
        for i2 in range(i0, i1):
            hgs = hg_loci[i2]
            for hg in hgs:
                if hg is not None:
#                    clan_arr[i2 - i0, ghg2i[g][hg]] = 1
                    clan_arr[i2 - i0, ghg2i[g][hg]] = True
        clan_arrs.append(clan_arr)


    blast_ids = defaultdict(dict)
    print('\t\t\t\tLoading alignment results', flush = True)
    if not os.path.isdir(f'{hlg_dir}algn/'):
        os.mkdir(f'{hlg_dir}algn/')
    jsons = collect_files(f'{hlg_dir}algn/', 'json')
    finished_hgs = set(int(os.path.basename(x).replace('.json', '')) \
                    for x in jsons)
    missing_hgs = set(blast_hash.keys()).difference(finished_hgs)
    if missing_hgs:
        blast_hash = {k: blast_hash[k] for k in list(missing_hgs)}
        with mp.get_context('forkserver').Pool(processes = cpus) as pool:
            pool.starmap(rnd2_gen_blastids_mp, 
                         ((hg, group_dict, hgx_dir, hlg_dir, minid) \
                           for hg, group_dict in blast_hash.items()))
            pool.close()
            pool.join()
    blast_ids = read_hlg_sim_jsons(f'{hlg_dir}algn/')

    # could be expedited by multiprocessing a writer and numba njit the math
    print('\t\t\t\tCalculating similarity', flush = True)
    if blast_ids:
        for g, v in enumerate(groups):
            i0, i1 = v
            clan_arr = clan_arrs[g]
            loc2loc_sim_mngr(clan_arr, finished_file, i0, loci, hg_loci,
                             0, blast_ids, min_hlg_id, simfun)
        # the file is now complete
        shutil.move(f'{hlg_dir}loci.adj.tmp.w', f'{hlg_dir}loci.adj.tmp')


def rnd1_loc2loc_mngr(
    db, clanI, loci, hgLoci, hg_dir, hgx_dir, hlg_dir, index,
    minid = 30, min_hlg_id = 0.15, simfun = overlap, min_overlap = 2,
    cpus = 1, Q = None
    ):

    finished_file = f'{hlg_dir}{clanI}.sadj.tmp'
    # is this clan already completed?
    if os.path.isfile(finished_file) or os.path.isfile(finished_file + '.r'):
        return
    # was this clan partially written?
    elif os.path.isfile(f'{finished_file}.w'):
        max_complete = open_prev_res(finished_file, index)
    # nothing completed
    else:
        max_complete = 0

    blast_hash = defaultdict(list)
    # only examine loci that have not been completed
    hg2i = {}
    for i in range(max_complete, len(loci)):
        locus = loci[i]
        hgs = hgLoci[i]
        for i1, hg in enumerate(hgs):
            if hg is not None:
                blast_hash[hg].append(locus[i1])
                if hg not in hg2i:
                    hg2i[hg] = len(hg2i)
    clan_arr = lil_matrix((len(loci) - max_complete, len(blast_hash)), dtype=bool)
    for i, v in enumerate(loci[max_complete:]):
        hgs = hgLoci[i + max_complete]
        for i1, hg in enumerate(hgs):
            if hg is not None:
                clan_arr[i, hg2i[hg]] = True


    blast_ids = defaultdict(dict)
    if cpus < 2:
        for hg, genes in blast_hash.items():
            gene_len = len(genes)
            if gene_len > 1:
                gene_set = set(genes)
                algn_base = f'{hgx_dir}{hg}.out'
                try:
                    blast_result = parse_blast(algn_base, gene_len, gene_set,
                                                minid)
                    if blast_result is not None:
                        blast_ids[hg] = blast_result                
                except FileNotFoundError:
                    pass

    else:
        print('\t\t\t\t\tLoading alignment results', flush = True)
        if not os.path.isdir(f'{hlg_dir}algn/'):
            os.mkdir(f'{hlg_dir}algn/')
        jsons = collect_files(f'{hlg_dir}algn/', 'json')
        finished_hgs = set(int(os.path.basename(x).replace('.json', '')) \
                        for x in jsons)
        missing_hgs = set(blast_hash.keys()).difference(finished_hgs)
        if missing_hgs:
            blast_hash = {k: blast_hash[k] for k in list(missing_hgs)}
            with mp.get_context('forkserver').Pool(processes = cpus) as pool:
                pool.starmap(gen_blastids_mp, 
                             ((hg, genes, hgx_dir, hlg_dir, minid) \
                               for hg, genes in blast_hash.items() \
                               if len(genes) > 1))
                pool.close()
                pool.join()
        blast_ids = read_hlg_sim_jsons(f'{hlg_dir}algn/')

    # could be expedited by multiprocessing a writer and numba njit the math
    if blast_ids:
        if cpus > 1:
            print('\t\t\t\t\tCalculating locus similarity', flush = True)
            Q.put(None)
        loc2loc_sim_mngr(clan_arr, finished_file, max_complete, loci, hgLoci,
                         index, blast_ids, min_hlg_id, simfun)

#        else:
 #           hlg_sim_mp_mngr(hits, f'{finished_file}.w', 
  #                          max_complete, loci, hgLoci, index, 
   #                         blast_ids, hlg_dir, f'{hlg_dir}mtrx/', min_hlg_id, 
    #                        simfun, chunkscale = 100, cpus = cpus)
        # begin appending to temporary writing output
        # the file is now complete
        shutil.move(f'{hlg_dir}{clanI}.sadj.tmp.w', f'{hlg_dir}{clanI}.sadj.tmp')
   #     if cpus > 1: # leave up to zipping 
#            jsons = collect_files(f'{hlg_dir}algn/', 'json')
 #           for j_f in jsons:
  #              os.remove(j_f)


def merge_clan_loci(preclan_loci, gene2hg, hgx2omes_path, hgx2dist_path, 
                    scaf2gene2i, merge_via_sim = False, 
                    min_branch_sim = 0.5, phylo = None):

    derep_loci = dereplicate_loci(preclan_loci, gene2hg, 
                                 hgx2omes_path, hgx2dist_path, scaf2gene2i,
                                 min_branch_sim, phylo)

    if merge_via_sim:
        vclan_loci = copy.deepcopy(derep_loci)
        out_clan_loci = defaultdict(lambda: defaultdict(list))
        outclanI = defaultdict(lambda: defaultdict(list))
        for clan, scaf_dict in derep_loci.items():
            for scaf, loci in scaf_dict.items():
                gene2i = scaf2gene2i[scaf]
                while True:
                    loc0, omes0 = loci[0]
                    try:
                        loc1, omes1 = loci[1]
                    except IndexError:
                        out_clan_loci[clan][scaf].append(loc0)
                        break
                    final_0 = gene2i[loc0[-1]]
                    start_1 = gene2i[loc1[0]]
                    if start_1 - final_0 == 1:
                        branch_sim = treecalcs.calc_branch_sim(phylo, omes0, omes1)
                        if branch_sim > min_branch_sim:
                            loci[1] = [loc0 + loc1, list(set(omes0).union(set(omes1)))]
                        else:
                            out_clan_loci[clan][scaf].append(loc0)
                    else:
                        out_clan_loci[clan][scaf].append(loc0)
                    loci.pop(0)
    else:
        out_clan_loci = {}
        for clan, scaf_dict in derep_loci.items():
           out_clan_loci[clan] = {}
           for scaf, loci in scaf_dict.items():
               out_clan_loci[clan][scaf] = [x[0] for x in loci]

    clan_loci = {clan: list(chain(*list(scaf_dict.values()))) \
                 for clan, scaf_dict in out_clan_loci.items()}

    hg_loci = {}
    for clan, loci in clan_loci.items():
        hg_loci[clan] = []
        for locus in loci:
            hg_loci[clan].append([])
            for gene in locus:
                try:
                    hg_loci[clan][-1].append(gene2hg[gene])
                except KeyError:
                    hg_loci[clan][-1].append(None)

    return clan_loci, hg_loci


def dereplicate_loci(preclan_loci, gene2hg,
                     hgx2omes_path, hgx2dist_path, scaf2gene2i,
                     min_branch_sim = 0.5, phylo = None):
    with open(hgx2dist_path, 'rb') as raw:
        hgx2dist = pickle.load(raw)
    with open(hgx2omes_path, 'rb') as raw:
        hgx2omes = pickle.load(raw)

    gene2hgx = defaultdict(list)
    out_clan_loci = defaultdict(lambda: defaultdict(list))
    for clan, scaf_dict in preclan_loci.items():
        for scaf, loci in scaf_dict.items():
            gene2i = scaf2gene2i[scaf]
            for i, loc_d in enumerate(loci):
                loc, hgx = loc_d[0], loc_d[1]
                loci[i].append(hgx2dist[hgx])
                loci[i].append(hgx2omes[hgx])
 #               loci[i].append([])
                [gene2hgx[x].append(hgx) for x in loc]
            # sort the scaffolds loci from highest to lowest TMD
            loci = sorted(loci, key = lambda x: x[-2], reverse = True)
            # give the remaining overlapping genes to the highest distance group
            # will need to add the new HGxs in the end
            final_loci, singletons_prep = [], []
            while loci:
                locd0 = loci[0]

                loc0 = set(locd0[0])
                overi = 1
                for i1, locd1 in enumerate(loci[1:]):
                    loc1 = set(locd1[0])
                    locIntersection = loc0.intersection(loc1)
                    if locIntersection:
                        # index from the back because some may have hgx still
                        dist1, omes1 = locd1[-2], locd1[-1]
                        # the loci are sorted from largest to lowest distribution, so
                        # we award the overlap to the largest TMD
                        loc_dif = loc1.difference(loc0)
                        # delete the initial version with overlap
                        loci.pop(i1+overi)
                        overi -= 1
                        if loc_dif:
     #                       if len(loc_dif) == 1:
      #                          singletons_prep.append([list(loc_dif)[0], dist1, omes1])
       #                         continue
                            new_locs = [sorted(loc_dif, key = lambda x: gene2i[x])]
                            break_i = True
                            # check for contuity, break-up noncontinuous loci
                            while new_locs:
                                new_loc = new_locs[0]
                                break_i = False
                                i_loc = [gene2i[x] for x in new_loc]
                                for i2, v in enumerate(i_loc[:-1]):
                                    if not i_loc[i2+1] - v == 1:
                                        break_i = i2 + 1
                                        break
                                # if there's a break, reiterate and add singletons
                                if break_i:
                                    n_loc0 = new_loc[:break_i]
                                    n_loc1 = new_loc[break_i:]
                      #              if len(n_loc0) > 1:
                                    new_locs.append(n_loc0)
                        #            else:
                         #               singletons_prep.append([n_loc0[0], dist1, omes1])
                          #          if len(n_loc1) > 1:
                                    new_locs.append(n_loc1)
                                    #else:
                                     #   singletons_prep.append([n_loc1[0], dist1, omes1])
                                else:
                                    loci.insert(i1+overi+1, [new_loc, dist1, omes1])
                                    overi += 1
                                new_locs.pop(0)

                if len(loc0) > 1:
                    final_loci.append(locd0)
                else:
                    singletons_prep.append([locd0[0][0], locd0[-2], locd0[-1]])
                loci.pop(0)
    
            genes_set = set(chain(*[x[0] for x in final_loci]))
            gene2loc_i, i2gene = {}, {}
            for i, locd in enumerate(final_loci):
                loc, omes = locd[0], locd[-1]
                for gene in loc:
                    gene2loc_i[gene] = [i, omes]
                    i2gene[gene2i[gene]] = gene

            #singletons_prep1 = [x for x in singletons_prep \
             #                   if x[0] not in genes_set]
            # sort the singletons by TMD, highest to lowest to take the largest
            # branch length representative of the singleton
            sorted_s_prep = sorted(singletons_prep,
                                key = lambda y: y[-2], reverse = True)
            singletons, s_set = [], set()
            for s in sorted_s_prep:
                if s[0] not in s_set:
                    singletons.append(s)
                    s_set.add(s[0])

            singletons = sorted(singletons, key = lambda y: gene2i[y[0]])
            todel = True
            while todel:
                todel = []
                for i, v in enumerate(singletons):
                    s, null, omes0  = v
                    s_i = gene2i[s]
                    t, b = s_i - 1, s_i + 1
                    if t in i2gene:
                        t_g = i2gene[t]
                        t_i, t_o = gene2loc_i[t_g]
                        t_sim = treecalcs.calc_branch_sim(phylo, omes0, t_o)
                    else:
                        t_sim = 0
                    if b in i2gene:
                        b_g = i2gene[b]
                        b_i, b_o = gene2loc_i[b_g]
                        b_sim = treecalcs.calc_branch_sim(phylo, omes0, b_o)
                    else:
                        b_sim = 0
                    # if there are both directions, award to the most similar
                    # distribution
                    if t_sim > min_branch_sim and b_sim > min_branch_sim:
                        if t_sim == b_sim:
                            merge_base = random.choice([0, 1])
                            if merge_base:
                                merge_i, merge_o = t_i, t_o
                            else:
                                merge_i, merge_o = b_i, b_o
                        elif t_sim > b_sim:
                            merge_i = t_i
                            merge_o = t_o
                        else:
                            merge_i = b_i
                            merge_o = b_o
                    elif t_sim > min_branch_sim:
                        merge_i = t_i
                        merge_o = t_o
                    elif b_sim > min_branch_sim:
                        merge_i = b_i
                        merge_o = b_o
                    else:
#                        continuities_prep.append([s, null, omes0])
                        continue
                    gene2loc_i[s] = [merge_i, merge_o]
                    i2gene[gene2i[s]] = s
                    final_loci[merge_i][0].append(s)
                    todel.append(i)
                for i in reversed(todel):
                    singletons.pop(i)
                    
            continuities_iz = [gene2i[x[0]] for x in singletons]
            continuities = []
            for i, v in enumerate(continuities_iz[:-1]):
                if continuities_iz[i + 1] - v == 1:
                    continuities.append(i)

            merges = []
            for i in continuities:
                s0, s1 = singletons[i], singletons[i+1]
                omes0 = s0[-1]
                omes1 = s1[-1]
                branch_sim = treecalcs.calc_branch_sim(phylo, omes0, omes1)
                if branch_sim > min_branch_sim:
                    merges.append([{s0[0], s1[0]}, set(omes0).union(set(omes1))])

            for i0, locd0 in enumerate(merges):
                loc0, omes0 = locd0
                try:
                    loc1, omes1 = merges[i0+1]
                except IndexError:
                    if len(loc0) == 1:
                        merges[i0] = None
                    continue
                locIntersection = loc0.intersection(loc1)
                if locIntersection:
                    merges[i0 + 1] = [loc0.union(loc1), 
                                      omes0.union(omes1)]
                merges[i0] = None

            for i, locd in enumerate(merges):
                if locd:
                    loc, omes = locd
                    final_loci.append([list(loc), list(omes)])

            for locd in final_loci:
                loc, omes = locd[0], locd[-1]
                out_clan_loci[clan][scaf].append([sorted(loc, key = lambda x: gene2i[x]),
                                                  omes])

    clan_loci = {} 
    for clan, scaf_dict in out_clan_loci.items():
        clan_loci[clan] = {}
        for scaf, loci in scaf_dict.items():
            clan_loci[clan][scaf] = []
            for loc, omes in loci:
                clan_loci[clan][scaf].append([loc, omes])
            clan_loci[clan][scaf] = sorted(clan_loci[clan][scaf],
                                           key = lambda x: scaf2gene2i[scaf][x[0][0]])

    return clan_loci #, out_hg_loci



def hash_clan_loci(ome, gff_path, ome_sig_clus, gene2hg, clusplusminus,
                   hgx2dist_path = None, hgx2omes_path = None,
                   merge_via_sim = False, min_branch_sim = 0.5, phylo = None):
    """sliding window of size clusplusminus, identify sequences that may have
    hgs that may be significant, grab their window, and check for a significant
    higher order og combo in the locus"""

    gff_list, protoclus, clus_out = gff2list(gff_path), {}, []
    cds_dict = input_parsing.compileCDS(gff_list, os.path.basename(gff_path).replace('.gff3',''))
    scaf2gene2i = {}
    for scaf, prots in cds_dict.items():
        scaf2gene2i[scaf] = {}
        for i, gene in enumerate(prots):
            scaf2gene2i[scaf][gene] = i

    preclan_loci = defaultdict(lambda: defaultdict(list))
    for scaf in cds_dict:
        for i0, seq0 in enumerate(cds_dict[scaf]):
            if seq0 in ome_sig_clus: # if the og is part of a significant seed
            # locus [[set(hgx)], i]
                for sig_clus in ome_sig_clus[seq0]:
                    clan = sig_clus[1]
                    start, end = None, None
                    if i0 < clusplusminus: # is i0 - clusplusminus < 0 ?
                        locus = cds_dict[scaf][:i0+clusplusminus+1] # then gather
                        # all the beginning
                    else:
                        locus = cds_dict[scaf][i0-clusplusminus:i0+clusplusminus+1]
                        # instead get the +/- and adjust for python
                    for i1, seq1 in enumerate(locus): # for each index and sequence
                    # in the locus
                        try:
                            og = gene2hg[seq1]
                        except KeyError:
                            continue
                        if og in sig_clus[0] and start is None: # if the og is
                        # in the sig clus and we haven't started
                            start = i1 # this is the start index
                        elif og in sig_clus[0]: # otherwise if it is in the
                        # sig clus label this the other border unless another is
                        # found
                            end = i1 + 1
                    hgx = tuple(sorted(sig_clus[0]))
                    preclan_loci[clan][scaf].append([
                         locus[start:end], hgx
                         ])

    out_clan_loci, out_hg_loci = merge_clan_loci(preclan_loci, gene2hg,
                                                 hgx2omes_path,
                                                 hgx2dist_path,
                                                 scaf2gene2i, merge_via_sim,
                                                 min_branch_sim, phylo)
    return ome, out_clan_loci, out_hg_loci


def read_to_write(out_h, in_adj, min_hlg):
    with open(in_adj, 'r') as raw:
        for line in raw:
            d = line.rstrip().split()
            hlg_id = round(float(d[2]) * 100)
            if hlg_id >= min_hlg:
                out_h.write(line.rstrip() + '\n')
    out_h.flush()


def write_adj_matrix(out_file, hlg_dir, min_loc_id):
    comp_hlg_id = round(float(min_loc_id) * 100)
    with open(out_file + '.tmp', 'w') as out:
        # get previously read adjacency matrices
        init_files = collect_files(hlg_dir, 'sadj.tmp.r')
        for f in init_files:
            read_to_write(out, f, comp_hlg_id)

        # get ready to read adjacency matrices
        files = collect_files(hlg_dir, 'sadj.tmp')
        f_set = set(files)

        # while the completion signal has not been received
        while f'{hlg_dir}done.sadj.tmp' not in f_set:
            for f in files:
                read_to_write(out, f, comp_hlg_id)
                shutil.move(f, f'{f}.r')
            files = collect_files(hlg_dir, 'sadj.tmp')
            f_set = set(files)

        # get final files
        for f in files:
            if f == f'{hlg_dir}done.sadj.tmp':
                os.remove(f'{hlg_dir}done.sadj.tmp')
            else:
                read_to_write(out, f, comp_hlg_id)
                shutil.move(f, f'{f}.r')
    shutil.move(out_file + '.tmp', out_file)
    for f in collect_files(hlg_dir, 'sadj.tmp.r'):
        os.remove(f)


def check_tune(tune, hlg_hgxs, hlg_omes):
    failed = []
    satisfied = False
    for name, data in tune.items():
        hgx, omes, false_omes = data[0], data[1], data[2]
        set_hgx = set(hgx)
        set_omes = set(omes)
        set_false = set(false_omes)
        check = [v for i, v in enumerate(hlg_hgxs) \
                 if set_hgx.issubset(set(v)) \
                 and set_omes.issubset(hlg_omes[i]) \
                 and not set_false.intersection(hlg_omes[i])]
        if check:
            continue
        else:
            failed.append(name)
    if failed:
        for name in failed:
            print(f'\t\t\t{name} was not recovered', flush = True)
    else:
        print(f'\t\t\tTuned successfully', flush = True)
        satisfied = True
    return satisfied


def read_clus(hlg_dir, mci2pre):
    t_hlgs = defaultdict(list)
    with open(hlg_dir + 'loci.clus', 'r') as raw:
        for line in raw: # loci indices
            hlg, mci = [int(x) for x in line.rstrip().split()]
            t_hlgs[hlg].append(mci2pre[mci])

    return [v for k, v in sorted(t_hlgs.items(), key = lambda x: x[0]) \
            if len(v) > 1]
  #          indices = [int(x) for x in line.rstrip().split()]
            # singletons won't pass thresholds, disregard them
    #        if len(indices) > 1:
     #           t_hlgs.append(indices)



def mcl2hlgs(hlg_dir, loci, hgxXloci, ome2i, inflation, 
             loc2clan, tune = False, min_omes = 1, cpus = 1):
    satisfied = False
    if not os.path.isfile(hlg_dir + 'loci.clus') or tune:
        print(f'\t\t\tRunning MCL - inflation: {inflation}', flush = True)
        if cpus > 5:
            mcl_threads = 10 # cap at 5 processors, ten threads
        else:
            mcl_threads = (cpus * 2) - 1
        MCL(hlg_dir + 'loci.adj', hlg_dir + 'loci.clus',
                      inflation = inflation, threads = mcl_threads)
        # could add iterative subsample MCL option here

    mci2pre = read_mci_rows(hlg_dir + 'mcl_rows.tsv')
    t_hlgs = read_clus(hlg_dir, mci2pre)

    # list(hlg) = [{hgx: (omes,)}]
    hlgs, hlg_hgxs, hlg_omes, hlg2dom = [], [], [], {}
    with open(f'{hlg_dir}loci.txt', 'w') as out:
        out.write('#omeI locI clanI prehlg locus\n')
        for hlg, locIs in enumerate(t_hlgs):
#            hlgs.append(defaultdict(list))
            hlgs.append([])
            hlg_hgxs.append([])
            hlgOme_list = [] 
            locs = []
            for locI in locIs:
                loc = loci[locI]
                locs.append((loc, locI))
                hgxs = hgxXloci[locI]
                # really should be done above to make hgxs formatted right
                omeI = ome2i[loc[0][:loc[0].find('_')]]
#                [hlgs[-1][hgx].append(omeI) for hgx in hgxs]
                hlgs[-1].append((loc, locI))
                hlgOme_list.append(omeI)
                hlg_hgxs[-1].extend(hgxs)
            if len(set(hlgOme_list)) > min_omes: # need more than 1
#                hlgs[-1] = {k: sorted(set(v)) for k,v in hlgs[-1].items()}
                if hlgs[-1]:
                    hlgs[-1] = [tuple(x) for x in hlgs[-1]]
                    hlg_hgxs[-1] = tuple(sorted(set(hlg_hgxs[-1])))
                    hlg_omes.append(tuple(sorted(set(hlgOme_list))))
#                    hlg_hgxs.append(tuple(sorted(set(chain(*list(hlgs[-1].keys()))))))
  #                  hlg_omes.append(tuple(sorted(set(chain(*list(hlgs[-1].values()))))))
                    prehlg = len(hlgs) - 1
                    hlg2dom[prehlg] = loc2clan[locs[0][1]]
                    for i, omeI in enumerate(hlgOme_list):
                        locus, locI = locs[i]
                        out.write(f'{omeI} {locI} {loc2clan[locI]} {prehlg} {",".join(locus)}\n')
                else:
                    del hlgs[-1]
                    del hlg_hgxs[-1]
            else:
                del hlgs[-1]
                del hlg_hgxs[-1]

    failed = []
    if tune:
        satisfied = check_tune(tune, hlg_hgxs, hlg_omes)
    else:
        satisfied = True

    list_hlg2dom = []
    for i in range(len(hlgs)):
        list_hlg2dom.append(hlg2dom[i])

    return satisfied, {i: v for i, v in enumerate(hlgs)}, \
           {i: v for i, v in enumerate(hlg_hgxs)}, \
           {i: v for i, v in enumerate(hlg_omes)}, \
           {i: v for i, v in enumerate(list_hlg2dom)}

def parse_and_sort_gff(gff_path):
    # order the loci in each scaffold
    cds_dict = input_parsing.compileCDS(gff2list(gff_path),
                                        os.path.basename(gff_path).replace('.gff3',''))
    scaf2gene2i = {}
    for scaf, prots in cds_dict.items():
        scaf2gene2i[scaf] = {}
        for i, gene in enumerate(prots):
            scaf2gene2i[scaf][gene] = i
   
    return scaf2gene2i, cds_dict

def hash_ome_lg_loc(gff_path, locs_y_lg, max_between = 2):
    scaf2gene2i, cds_dict = parse_and_sort_gff(gff_path)
    scaf2i2gene = {}
    for scaf, gene2i in scaf2gene2i.items():
        scaf2i2gene[scaf] = {v: k for k, v in gene2i.items()}
    scaf2locs, null = output_data.clus2scaf(cds_dict, locs_y_lg)

    hlg2scaf2locs = defaultdict(lambda: defaultdict(list))
    for scaf, locs in scaf2locs.items():
        for loc, hlg in locs:
            hlg2scaf2locs[hlg][scaf].append(loc)
    
    hlg2locs = {}
    for hlg, scaf_dict in hlg2scaf2locs.items():
        hlg2locs[hlg] = []
        for scaf, locs in scaf_dict.items():
            while len(locs) > 1:
                loc0 = locs[0]
                loc1 = locs[1]
                top_i = scaf2gene2i[scaf][loc0[-1]]
                bot_i = scaf2gene2i[scaf][loc1[0]]
                if bot_i - top_i == 1:
                    locs[0].extend(locs[1])
                    locs.pop(1)
                # award multiple intervening genes if they meet the max_between criterium
                elif bot_i - top_i == max_between + 1:
                    locs[0].extend([scaf2i2gene[scaf][top_i + bw_i] \
                                    for bw_i in range(1, max_between + 1)] + locs[1])
                    locs.pop(1)
                else:
                    hlg2locs[hlg].append(loc0)
                    locs.pop(0)
            hlg2locs[hlg].append(locs[0])

    return hlg2locs


def lg_loc_mngr(db, lgs, gene2hg, max_between = 2, cpus = 1):
    omes2loci = defaultdict(list)
    for lg, locs in lgs.items():
        for loc, locI in locs:
            ome = loc[0][:loc[0].find('_')]
            omes2loci[ome].append((loc, lg))

    with mp.get_context('forkserver').Pool(processes = cpus) as pool:
        lg_loci = pool.starmap(hash_ome_lg_loc,
                               tqdm(((db[ome]['gff3'], omes2loci[ome],
                                      max_between) for ome in db.keys()), 
                                    total = len(db)))

    lg2locs = defaultdict(list)
    lg2hg_locs = defaultdict(list)
    lg2hgxs_locs = defaultdict(list)
    for res in lg_loci:
        for lg, locs in res.items():
            for loc in locs:
                hg_loc = []
                for gene in loc:
                    try:
                        hg_loc.append(gene2hg[gene])
                    except KeyError:
                        hg_loc.append(None)
                lg2locs[lg].append(loc)
                lg2hg_locs[lg].append(hg_loc)

    return lg2locs, lg2hg_locs    
#            db, clanI, locs, clan_hg_loci[clanI],
 #           hg_dir, hgx_dir, hlg_dir, index, minid, min_loc_id,
  #          simfun, min_overlap

def refine_group(db, ome2i, group_hg_loci, group_loci,
                 hg_dir, hgx_dir, wrk_dir, grp_dir, minid, min_loc_id,
                 simfun, inflation, tune, min_omes, prefix, cpus = 1, rnd = 1):

    m, adj_mtr = mp.Manager(), grp_dir + 'loci.adj'
    if cpus < 2:
        cpus = 2

#    min_overlap = round((min_loc_id * 2) - 0.5)
    min_overlap = 2 # need at least two genes of overlap to call it a cluster
    index, cmds = 0, []

    protog2h = {gI: Counter(chain(*hg_locs)) \
                for gI, hg_locs in enumerate(group_hg_loci)}
    group2hgx = {gI: set([k for k, v in h.items() if v > 1]) \
                 for gI, h in protog2h.items()}

    if rnd != 2:
        loci, hgxXloci, loc2group = {}, {}, {}
        for gI, locs in enumerate(group_loci):
            if not os.path.isfile(f'{grp_dir}{gI}.sadj.tmp.r'):
                cmds.append([
                    db, gI, locs, group_hg_loci[gI],
                    hg_dir, hgx_dir, grp_dir, index, minid, min_loc_id,
                    simfun, min_overlap
                    ])
            for i, locus in enumerate(locs):
                hg_loc = set(group_hg_loci[gI][i])
                loci[index] = group_loci[gI][i]
                hgxXloci[index] = tuple(sorted([x for x in \
                                       list(hg_loc.intersection(group2hgx[gI])) \
                                       if x is not None]))
                loc2group[index] = gI
                index += 1
        if not os.path.isfile(adj_mtr):
            W = mp.Process(target = write_adj_matrix, 
                           args = (adj_mtr, grp_dir, min_loc_id))
            W.start()
    #        print(f'\t\t\t{max([len(x) for x in )} loci in clan 0', flush = True)
            procs = []

            if cmds:
                if len(cmds[0][2]) > 30000 and cmds[0][1] == 0:
                    print(f'\t\t\t\tRunning group 0', flush = True)
                    Q = mp.Manager().Queue()
                    procs = [mp.Process(target=rnd1_loc2loc_mngr,
                                    args=(*cmds[0], cpus - 1, Q))]
                    procs[0].start()
                    Q.get()
                    print('\t\t\t\tRunning all groups', flush = True)
                    del cmds[0]
                with mp.get_context('forkserver').Pool(processes = cpus - 1) as pool:
                    pool.starmap(rnd1_loc2loc_mngr, tqdm(cmds, total = len(cmds), miniters = 1))
                    pool.close()
                    pool.join()
                if procs:
                    procs[0].join()
                    procs[0].close()
            with open(f'{grp_dir}done.sadj.tmp', 'w') as out:
                pass
            W.join()
    else:
        index, hgxXloci, loc2group, loci, hg_loci, groups = 0, {}, {}, {}, {}, []
        for gI, locs in enumerate(group_loci):
            groups.append([index, -1])
            for i, locus in enumerate(locs):
                hgs = group_hg_loci[gI][i]
                hg_loci[index] = hgs
                hg_loc = set(hgs)
                loci[index] = group_loci[gI][i]
                hgxXloci[index] = tuple(sorted([x for x in \
                                       list(hg_loc.intersection(group2hgx[gI])) \
                                       if x is not None]))
                loc2group[index] = gI
                index += 1
            groups[-1][1] = index
        if not os.path.isfile(f'{grp_dir}loci.adj.tmp') \
            and not os.path.isfile(f'{grp_dir}loci.adj'):
            rnd2_loc2loc_mngr(grp_dir, loci, hg_loci, hgx_dir, 
                              groups, minid, min_loc_id, simfun, cpus)
            os.rename(f'{grp_dir}loci.adj.tmp', f'{grp_dir}loci.adj')

    satisfied = False
    if tune:
        inflation = 1.1
    while not satisfied:
        if tune:
            inflation += 0.1
            if inflation > 3:
                break
        satisfied, grps, grp_hgxs, grp_omes, lilg2bigg = mcl2hlgs(grp_dir, loci,
                                                           hgxXloci, ome2i, 
                                                           inflation, loc2group,
                                                           tune, min_omes,
                                                           cpus)
    if tune:
        if not satisfied:
            eprint('\nERROR: tuning failed to recover clusters', flush = True)
            sys.exit(135)
        inf_strp = str(inflation * 10)[:2]
        inf_str = inf_strp[0] + '.' + inf_strp[1]
        with open(f'{wrk_dir}../log.txt', 'r') as raw:
            d = raw.read().replace(f'{prefix}_inflation\tNone', f'{prefix}_inflation\t' + inf_str)
        with open(f'{wrk_dir}../log.txt', 'w') as out:
            out.write(d)

    return grps, grp_hgxs, grp_omes, lilg2bigg, hgxXloci


def comp_scaf2hgs(cds_dict, gene2hg):
    scaf2hgs = {}
    for scaf, prots in cds_dict.items():
        scaf2hgs[scaf] = []
        for prot in prots:
            try:
                scaf2hgs[scaf].append(gene2hg[prot])
            except KeyError:
                scaf2hgs[scaf].append(None)
    return scaf2hgs


def extend_loci(ome, gff_path, loc2need, gene2hg, forbid_genes, 
                ext_len = 1, min_new_cont = 2, announcements = {87332}):
    forbidden_genes = set(forbid_genes)
    scaf2gene2i, cds_dict = parse_and_sort_gff(gff_path)
    
    gene2scaf, index_map = {}, {}
    for scaf, genes in cds_dict.items():
        index_map[scaf] = {v: i for i, v in enumerate(genes)}
        for gene in genes:
            gene2scaf[gene] = scaf

    # currently only built for an extension length of 1, will need modification
    # to accomodate greater lengths, e.g. determining how to merge an extension
    # if only the nearest-to-the-cluster entry of the extension meets missing_hg
    mod_locs, edge_locs = {}, {}
    for locI, d in loc2need.items():
        # opt to declare contig edges based on the starting locus, not the end
        # this is a decision that has arguments for/against both mechanisms
        # I'm setting it on the start to be conservative about this module
        contig_edge = False
        # keep track of modified loci, or contig edge loci for future mod
        modified = False
        loc, hlg, miss_hg_tup = d

        # announce this GCF for auditing
#        if gcf in announcements:
 #           announce = True
  #      else:
   #         announce = False

        miss_hg = set(miss_hg_tup)
        gene0 = loc[0]
        scaf = gene2scaf[gene0]
        scaf_len = len(cds_dict[scaf])
        ord_loc = sorted(loc, key = lambda pair: index_map[scaf][pair])
        # identify whether the top/bottom can be extended
        bot_i, top_i = scaf2gene2i[scaf][ord_loc[0]], scaf2gene2i[scaf][ord_loc[-1]]
        if bot_i:
            bot_i -= ext_len
            bot_ex = cds_dict[scaf][bot_i]
            # do NOT interfere with existing clusters
            if bot_ex in forbidden_genes:
                bot_ex = False
        else:
            contig_edge = True
            bot_ex = False
        if top_i < scaf_len - 1:
            top_i += ext_len
            top_ex = cds_dict[scaf][top_i]
            if top_ex in forbidden_genes:
                top_ex = False
        else:
            contig_edge = True
            top_ex = False

        # while there is a true extension possibility
        while top_ex or bot_ex:
            todel = set()
            if top_ex:
                try:
                    top_hg = gene2hg[top_ex]
                except KeyError:
                    top_hg = None
                if top_hg in miss_hg:
                    modified = True
                    ord_loc.append(top_ex)
                    todel.add(top_hg)
                    if top_i < scaf_len - 1:
                        top_i += ext_len
                        top_ex = cds_dict[scaf][top_i]
                        if top_ex in forbidden_genes:
                            top_ex = False
                else:
                    top_ex = False
            if bot_ex:
                try:
                    bot_hg = gene2hg[bot_ex]
                except KeyError:
                    bot_hg = None
                if bot_hg in miss_hg:
                    modified = True
                    ord_loc.insert(0, bot_ex)
                    todel.add(bot_hg)
                    if bot_i:
                        bot_i -= ext_len
                        bot_ex = cds_dict[scaf][bot_i]
                        if bot_ex in forbidden_genes:
                            bot_ex = False
                else:
                    bot_ex = False
            # dont edit the miss_hg until the end to not bias for/against top/bot    
            miss_hg = miss_hg.difference(todel)
        if modified:
            mod_locs[locI] = (ord_loc, hlg)
        
        # its risky to extend to different contigs based on a single HG, so we
        # should only do it if there's evidence that exceeds min_new_cont (2)
        if contig_edge and len(miss_hg) >= min_new_cont:
            edge_locs[locI] = (ord_loc, hlg, miss_hg)
        
    # if there are loci at the edge of a contig
    to_add_locs = []
    if edge_locs:
        scaf2hgs = comp_scaf2hgs(cds_dict, gene2hg)
        # we are only looking for extensions of min_new_cont length (2)
        # anything greater than this should have been picked up by a different
        # mechanism
        for locI, d in edge_locs.items():
            loc, hlg, miss_hg = d

            gene0 = loc[0]
            loc_scaf = gene2scaf[gene0]
            found = False
            # it will extend based on the first scaffold that meets the criteria
            for scaf, hg_list in scaf2hgs.items():
                # ignore self-scaffold check
                if loc_scaf == scaf:
                    continue
                bot_prot = cds_dict[scaf][:min_new_cont]
                top_prot = cds_dict[scaf][-min_new_cont:]
                # ignore other cluster entries
                if all(x not in forbidden_genes for x in bot_prot):
                    # this approach allows the rare situation for both sides of 
                    # the contig to meet the criteria before breaking
                    bot_set = hg_list[:min_new_cont]
                    if len(set(bot_set).intersection(miss_hg)) >= min_new_cont:
                        to_add_locs.append((bot_prot, hlg))
                        found = True
                else:
                    bot_set = set()
                if all(x not in forbidden_genes for x in top_prot):
                    top_set = hg_list[-min_new_cont:]
                    if len(set(top_set).intersection(miss_hg)) >= min_new_cont:
                        if top_set != bot_set:
                            to_add_locs.append((top_prot, hlg))
                            found = True
                if found:
                    break

    # ideally would like to propagate contig edge to downstream gbk because
    # antiSMASH does it and software like BiG-SCAPE use that data
    return ome, mod_locs, to_add_locs


def extend_loc_mngr(db, hlgs, hlg_hgxs, ome2gene2hg, hgxXloci, hlg_dir,
                    ext_len = 1, min_new_cont = 2, cpus = 1):
    ome2genes = defaultdict(list)
    ome2loc2need = defaultdict(dict)
    for hlg, locs in hlgs.items():
        hlg_hg_set = set(hlg_hgxs[hlg])
        for loc, locI in locs:
            ome = loc[0][:loc[0].find('_')]
            hg_loc = hgxXloci[locI]
            miss_hg = hlg_hg_set.difference(set(hg_loc))
            if miss_hg:
                ome2loc2need[ome][locI] = (loc, hlg, tuple(miss_hg))
            ome2genes[ome].extend(loc)
   
    with mp.Pool(processes = cpus) as pool:
        modifications = pool.starmap(extend_loci, 
                                     ((ome, db[ome]['gff3'], loc2need,
                                       ome2gene2hg[ome], ome2genes[ome],
                                       ext_len, min_new_cont) \
                                      for ome, loc2need in ome2loc2need.items()))

    with open(f'{hlg_dir}edge_loci.tsv', 'w') as edge_out:
        edge_out.write('#hlg\tlocus\n')
        with open(f'{hlg_dir}extension_loci.tsv', 'w') as extension_out:
            extension_out.write('#hlg\toriginal_locus\tupdate_locus\n')
            mod_count, cont_count = 0, 0
            for ome, mod_locs, to_add_locs in modifications:
                mod_count += len(mod_locs)
                for locI, d in mod_locs.items():
                    loc, hlg = d
                    ome2genes[ome].extend(loc)
                    for i, v in enumerate(hlgs[hlg]):
                        t_loc, t_locI = v
                        if t_locI == locI:
                            extension_out.write(f'{hlg}\t{",".join(t_loc)}\t{",".join(loc)}\n')
                            hlgs[hlg][i] = (loc, None)
                ome2genes[ome] = set(ome2genes[ome])
                for loc, hlg in to_add_locs:
                    # this awards the new loci to the first entry, which doesn't
                    # have an order that is necessarily meaningful
                    if all(x not in ome2genes[ome] for x in loc):
                        cont_count += 1
                        hlgs[hlg].append((loc, None))
                        edge_out.write(f'{hlg}\t{",".join(loc)}\n')
                        ome2genes[ome] = ome2genes[ome].union(set(loc))

        
    # we aren't modifying the hgs of each hlg so can ignore them
    return {hlg: tuple([loc[0] for loc in locs]) for hlg, locs in hlgs.items()}, \
           mod_count, cont_count


def classify_hlgs(
    hgx2loc, db, gene2hg, i2hgx, hgx2i,
    phylo, ome2i, hgx2omes, hg_dir, hgx_dir,
    wrk_dir, ome2partition, bord_scores_list,
    hg2gene, omes2dist = {}, clusplusminus = 3, inflation_1 = 1.1,
    inflation_2 = 1.5, min_hgx_overlap = 1, min_omes = 2, 
    minid = 30, cpus = 1, algorithm = 'diamond',
    min_loc_id = 0.3, simfun = overlap, printexit = False,
    tune = False, dist_func = treecalcs.calc_tmd,
    uniq_sp = False, min_branch_sim = 0.8, algn_sens = '',
    skipalgn = False, fallback = False, merge_via_sim = False,
    max_between = 2
    ):

    i2ome = {v: k for k, v in ome2i.items()}

    if not checkdir(hgx_dir, unzip = True, rm = True):
        os.mkdir(hgx_dir)

    groupI = wrk_dir + 'gene2hgx.pickle'
    groupII = wrk_dir + 'hgcs.pickle'

    # make an ome-centric hash for less MP overhead
    ome2gene2hg = defaultdict(dict)
    for gene, hg in gene2hg.items():
        ome = gene[:gene.find('_')]
        ome2gene2hg[ome][gene] = hg

    if not os.path.isfile(groupI):
        print('\tAggregating HGxs with overlapping loci', flush = True)
        hgx_genes = {}
        # compile a hash that shows the HGx each gene corresponds to
        for hgx in hgx2loc:
            if hgx in hgx2i: # if it passed the border threshold
                for gene in hgx2loc[hgx]:
                    ome = gene[:gene.find('_')]
                    if ome not in hgx_genes:
                        hgx_genes[ome] = {}
                    if gene not in hgx_genes[ome]:
                        hgx_genes[ome][gene] = []
                    hgx_genes[ome][gene].append(hgx)

        hashOgx_cmds = []
        start = datetime.now()
        for ome in hgx_genes:
            hashOgx_cmds.append([
                db[ome]['gff3'],
                ome, hgx_genes[ome], ome2gene2hg[ome], clusplusminus
                ])
        # expand the hgx 2 gene hash to include all genes in the locus corresponding
        # to the hgx
        with mp.get_context('fork').Pool(processes = cpus) as pool:
            gene_tups = pool.starmap(hash_hgx, hashOgx_cmds)
            pool.close()
            pool.join()    

        gene2hgx, hgx2genes = {}, {}

        for res in gene_tups:
            gene2hgx[ome2i[res[0]]] = {x[0]: set(x[1]) for x in res[1]}
            for hgx, hgs in res[2]:
                if hgx not in hgx2genes:
                    hgx2genes[hgx] = {og: [] for og in hgx}
                for og, seqs in hgs:
                    hgx2genes[hgx][og].extend(seqs)
            # hgx2genes = {hgx: og: [gene, gene]}
            # gene2hgx = {omeI: gene: {HGx}}

        with open(groupI, 'wb') as out:
            pickle.dump([gene2hgx, hgx2genes], out)
        print('\t\t' + str(datetime.now() - start))
    else:
        print('\tLoading aggregated HGxs', flush = True)
        with open(groupI, 'rb') as in_:
            gene2hgx, hgx2genes = pickle.load(in_)


    if not os.path.isfile(groupII):
        print(f'\tIdentifying HGxs with at least {min_hgx_overlap} gene overlap', 
              flush = True)
        tpairDict = defaultdict(list)
        for ome, gene2hgx_ome in tqdm(gene2hgx.items(),
                                      total = len(gene2hgx)):
            tpairDict = find_hgx_pairs(
                gene2hgx_ome, ome, hgx2i, tpairDict,
                min_hgx_overlap = min_hgx_overlap
                )

        pairDict = {}
        for id_, omes in tpairDict.items():
            omes_set = set(omes)
   #         if len(omes_set) > 1: # need at least one ome to calc a branch length
            pairDict[id_] = tuple(sorted(omes_set))
#        omes2dist = treecalcs.update_dists(phylo, pairDict, cpus = cpus, omes2dist = omes2dist,
 #                                          func = dist_func, uniq_sp = uniq_sp, i2ome = i2ome)


        print('\tAggregating HGxs', flush = True)
        # populate a lil_matrix here, then use that network to identify modules
        print('\t\tBuilding binary HGx-HGx network', flush = True)
        matrix = lil_matrix((len(i2hgx), len(i2hgx)), dtype=bool)
        for idPair, omes in tqdm(pairDict.items(), total = len(pairDict)):
            i0, i1 = idPair[0], idPair[1]
            matrix[i0, i1] = True
            matrix[i1, i0] = True
        print('\t\tIsolating subgraphs (HGx clans)', flush = True)
        g = Graph(directed = False)
        idx = matrix.nonzero()
        g.add_edge_list(np.transpose(matrix.nonzero()))
        components = label_components(g)[0]
        subgraph_is = sorted(set(components.a))
#        network = nx.from_scipy_sparse_matrix(matrix)
 #       subgraphs = [network.subgraph(c) for c in nx.connected_components(network)]
        # use .copy() after iterator to copy
        clans, clanOmes, clanHGxs = [], [], []
#        for sg in subgraphs:
        for sgi in subgraph_is:
            sg = GraphView(g, vfilt = components.a == sgi)
            clans.append({})
            clanOmes.append([])
            clanHGxs.append([])
#            for id_ in list(sg.nodes):
            for id_ in sg.vertices():
                hgx = i2hgx[int(id_)]
                omes = hgx2omes[hgx]
                clans[-1][hgx] = omes # storing in a hash to allow future
                # manipulation
                clanHGxs[-1].extend(hgx)
                clanOmes[-1].extend(omes)
            clanOmes[-1] = tuple(sorted(set(clanOmes[-1])))
            clanHGxs[-1] = tuple(sorted(set(clanHGxs[-1])))

        omes2dist = treecalcs.update_dists(
            phylo, {clanHGxs[i]: omes for i, omes in enumerate(clanOmes)},
            cpus = cpus, omes2dist = omes2dist, func = dist_func, uniq_sp = uniq_sp,
            i2ome = i2ome
            )
        with open(groupII, 'wb') as out:
            pickle.dump([clans, clanOmes, clanHGxs], out)
    else:
        print('\tLoading HGx clans', flush = True)
        with open(groupII, 'rb') as in_:
            clans, clanOmes, clanHGxs = pickle.load(in_)

    print('\t\t' + str(len(clans)) + ' clans', flush = True)
    print('\tClassifying HGxs into locus domains', flush = True)
    clanFile = wrk_dir + 'clan2loci.'
    if not os.path.isfile(clanFile + 'hg.json.gz'):
        print('\t\tPreparing locus extraction', flush = True)
        ome2clus2extract = defaultdict(dict)
        for i, clan in enumerate(clans):
            for hgx, omes in clan.items():
                # create a seed hash of the hgx and its seed gene
                for gene in hgx2loc[hgx]:
                    ome = gene[:gene.find('_')]
                    if gene not in ome2clus2extract[ome]:
                        ome2clus2extract[ome][gene] = [[set(hgx), i]]
                    else:
                        ome2clus2extract[ome][gene].append([set(hgx), i])
                    # {ome: {gene: [[set(hgx), clanI]]}}

        print('\t\tExtracting clan loci', flush = True)
        with mp.get_context('forkserver').Pool(processes = cpus) as pool:
            hash_res = pool.starmap(hash_clan_loci, 
                                    tqdm(((ome, db[ome]['gff3'], clus2extract,
                                          ome2gene2hg[ome], clusplusminus,
                                          f'{wrk_dir}hgx2dist.pickle', 
                                          f'{wrk_dir}hgx2omes.pickle', merge_via_sim, min_branch_sim,
                                          phylo) \
                                          for ome, clus2extract \
                                          in ome2clus2extract.items()),
                                         total = len(ome2clus2extract)))
            pool.close()
            pool.join()

        print('\t\tWriting clan loci', flush = True)
        clan_loci = defaultdict(list)
#        clan_hgx4gcfs = defaultdict(list)
        clan_hg_loci = defaultdict(list)
        for ome, out_clan_loci, out_clan_hg_loci in hash_res:
            for clan in out_clan_loci:
                clan_loci[clan].extend(out_clan_loci[clan])
 #               clan_hgx4gcfs[clan].extend(out_clan_hgx[clan])
                clan_hg_loci[clan].extend(out_clan_hg_loci[clan])

        write_json(clan_loci, clanFile + 'json.gz')
  #      write_json(clan_hgx4gcfs, clanFile + 'hgx.json.gz')
        write_json(clan_hg_loci, clanFile + 'hg.json.gz')
    else:
        print('\t\tLoading clan loci', flush = True)
        clan_loci = {int(k): tuple(v) for k,v in sorted(
                read_json(clanFile + 'json.gz').items(), 
                key = lambda x: len(x[1]), reverse = True)}
   #     clan_hgx4gcfs = {int(k): tuple([tuple(i) for i in v]) \
    #                    for k,v in read_json(clanFile + 'hgx.json.gz').items()}
        clan_hg_loci = {int(k): tuple(v) \
                      for k, v in read_json(clanFile + 'hg.json.gz').items()}

    # dict(clan_loci) = {int(clanI): ((ome0_gene0, ome0_gene1,)...,)}
    # dict(clan_hgx4gcfs) = {int(clanI): ((og0, og1, ogn,)...,)}
    # dict(clan_hg_loci) = {int(clanI): ((ome0_gene0_og, ome0_gene1_og,)...,)}

    # need to move up to before writing
    clan_loci = {k: v for k,v in sorted(clan_loci.items(), key = lambda x: len(x[1]), reverse = True)}
    #clan_hgx4gcfs = [clan_hgx4gcfs[k] for k in clan_loci]
    clan_hg_loci = [clan_hg_loci[k] for k in clan_loci]
    clan_loci = [v for v in clan_loci.values()]

    hlg_dir = wrk_dir + 'hlg/' 
    if not os.path.isdir(hlg_dir):
        os.mkdir(hlg_dir)
    lg_dir = hlg_dir + 'lg/'
    if not os.path.isdir(lg_dir):
        os.mkdir(lg_dir)

    print('\t\tOutputting HG fastas', flush = True)
#    hgs_prep = set(chain(*list(chain(*list(clan_hg_loci.values())))))
    hgs_prep = list(chain(*list(chain(*clan_hg_loci))))
    hg_count = Counter(hgs_prep)
    if None in hg_count:
        del hg_count[None]
    hgs = sorted([x for x, v in hg_count.items() \
                  if v > 1])
    hg_dir = input_parsing.hg_fa_mngr(wrk_dir, None, 
                             hgs, db, hg2gene, cpus = cpus)
    evo_conco.run_blast(hgs, db, hg_dir, hgx_dir,
                        algorithm = algorithm, printexit = printexit,
                        sensitivity = algn_sens, hg2gene = hg2gene,
                        skipalgn = skipalgn, #minid = minid, let minid be at locus step
                        fallback = fallback, cpus = cpus)

    print('\t\tCalling locus domains (sub-cluster domains)', flush = True)
    clan_lens = [len(v) for v in clan_loci]
#    print('\t\t\t' + str(sum([len(v) for v in list(clan_loci.values())])) \
    print(f'\t\t\t{sum(clan_lens)}' \
        + ' loci', flush = True)
    print(f'\t\t\t{max(clan_lens)}' \
        + ' loci in clan 0', flush = True)
    lgs, lg_hgxs, lg_omes, lg2clan, hgxXloci = refine_group(
                 db, ome2i, clan_hg_loci, clan_loci,
                 hg_dir, hgx_dir, wrk_dir, lg_dir, minid, min_loc_id,
                 simfun, inflation_1, tune, min_omes, 'domain', cpus = cpus)
    print('\t\t\t' + str(len(lgs)) + ' domains', flush = True)

    print('\t\tMerging loci', flush = True)
    lg2locs, lg2hg_locs = lg_loc_mngr(db, lgs, gene2hg, max_between, cpus)
    lg_locs = {k: v for k,v in sorted(lg2locs.items(), key = lambda x: len(x[1]), reverse = True)}
    lg_hg_loci = [lg2hg_locs[k] for k in lg_locs]
    lg_locs = [v for v in lg_locs.values()]

    print('\t\tCalling HLGs (unfiltered-GCFs)', flush = True)
    lg_lens = [len(v) for v in lg_locs]
#    print('\t\t\t' + str(sum([len(v) for v in list(clan_loci.values())])) \
    print(f'\t\t\t{sum(lg_lens)}' \
        + ' loci', flush = True)
    print(f'\t\t\t{max(lg_lens)}' \
        + ' loci in LG 0', flush = True)
    hlgs, hlg_hgxs, hlg_omes, hlg2dom, hgxXloci = refine_group(
                 db, ome2i, lg_hg_loci, lg_locs,
                 hg_dir, hgx_dir, wrk_dir, hlg_dir, minid, min_loc_id,
                 sorensen, inflation_2, tune, min_omes, 'hlg', cpus = cpus, rnd = 2)
    print('\t\t\t' + str(len(hlgs)) + ' HLGs', flush = True)

    print('\tExtending incomplete loci and contig edge loci', flush = True)
    hlgs, mod_count, cont_count = extend_loc_mngr(db, hlgs, hlg_hgxs, ome2gene2hg, 
                                                  hgxXloci, hlg_dir, ext_len = 1,
                                                  min_new_cont = 2, cpus = cpus)
    print(f'\t\t{mod_count} loci extended', flush = True)
    print(f'\t\t{cont_count} loci merged from contig edges', flush = True)

    omes2dist = treecalcs.update_dists(
        phylo, {hlg_hgxs[i]: omes for i, omes in hlg_omes.items()},
        cpus = cpus, omes2dist = omes2dist, func = dist_func, 
        uniq_sp = uniq_sp, i2ome = i2ome
        )

    return hlgs, hlg_hgxs, hlg_omes, hlg2dom, omes2dist



import re
import os
import sys
import gzip
import shutil
import hashlib
import multiprocessing as mp
from tqdm import tqdm
from ast import literal_eval
from cogent3 import PhyloNode, load_tree
from collections import defaultdict
from itertools import combinations
from datetime import datetime
from mycotools.acc2fa import dbmain as acc2fa
from mycotools.lib.biotools import gff2list, dict2fa, fa2dict
from mycotools.lib.kontools import format_path, eprint, collect_files, collect_dirs

def init_log(
    log_file, log_dict
    ):
    with open(log_file, 'w') as out:
        out.write(f'mtdb\t{log_dict["mtdb"]}\n' \
                + f'focal_genes\t{log_dict["focal_genes"]}\n' \
                + f'microsynt_constraint\t{log_dict["microsynt_constraint"]}\n' \
                + f'microsynt_tree\t{log_dict["microsynt_tree"]}\n'
                + f'hg_file\t{log_dict["hg_file"]}\n' \
                + f'dist_type\t{log_dict["dist_type"]}\n' \
                + f'uniq_sp\t{log_dict["uniq_sp"]}\n' \
                + f'plusminus\t{log_dict["plusminus"]}\n' \
                + f'hgp_percentile\t{log_dict["hgp_percentile"]}\n' \
                + f'hgx_percentile\t{log_dict["hgx_percentile"]}\n' \
                + f'aligner\t{log_dict["aligner"]}\n' \
#                + f'hlg_percentile\t{log_dict["hlg_percentile"]}\n' \
                + f'gene_id\t{log_dict["gene_id"]}\n' \
                + f'hlg_id\t{log_dict["hlg_id"]}\n' \
                + f'hlg_sim\t{log_dict["hlg_sim"]}\n' \
                + f'orig_hlg_id\t{log_dict["hlg_id"]}\n' \
                + f'topology_merge\t{log_dict["topology_merge"]}\n' \
                + f'min_topology_sim\t{log_dict["min_topology_sim"]}\n' \
                + f'domain_inflation\t{log_dict["domain_inflation"]}\n' \
                + f'hlg_inflation\t{log_dict["hlg_inflation"]}\n' \
                + f'tuning\t{log_dict["tuning"]}\n' \
                + f'id_percent\t{log_dict["id_percent"]}\n' \
                + f'pos_percent\t{log_dict["pos_percent"]}\n' \
                + f'iipr_percent\t{log_dict["iipr_percent"]}\n' \
                + f'patch_threshold\t{log_dict["patch_threshold"]}\n' \
                + f'gcl_threshold\t{log_dict["gcl_threshold"]}\n' \
                + f'dist_threshold\t{log_dict["dist_threshold"]}\n' \
                + f'partition\t{log_dict["partition"]}\n' \
                + f'null_samples\t{log_dict["null_samples"]}\n' \
                + f'n50\t{log_dict["n50"]}\n' \
                + f'hg_dir\t{log_dict["hg_dir"]}\n' \
                + f'hgx_dir\t{log_dict["hgx_dir"]}\n' \
                + f'ipr_path\t{log_dict["ipr_path"]}\n' \
                + f'pfam_path\t{log_dict["pfam_path"]}')

def read_log(
    log_file, log_dict
    ):
    log_res = {}
    with open(log_file, 'r') as raw:
        for line in raw:
            key = line[:line.find('\t')]
            res = line[line.find('\t') + 1:].rstrip()
            if key not in log_dict: # old feature
                log_res[key] = False
                continue
            elif key == 'orig_hlg_id':
                if round(float(res) * 100) \
                   <= round(float(log_dict['hlg_id']) * 100):
                   log_res[key] = True
                else:
                   log_dict['orig_hlg_id'] = log_dict['hlg_id']
                   log_res[key] = False
            elif key == 'hgx_percentile':
                if round(float(res) * 100) > \
                    round(float(log_dict['hgx_percentile']) * 100):
                    log_res['hgx_percentile_I'] = False
                    log_res['hgx_percentile_II'] = False
                elif round(float(res) * 100) != round(float(log_dict['hgx_percentile']) * 100):
                    log_res['hgx_percentile_I'] = True
                    log_res['hgx_percentile_II'] = False
                else:
                    log_res['hgx_percentile_I'] = True
                    log_res['hgx_percentile_II'] = True
            elif key == 'hlg_id':
                continue
            elif key == 'hlg_sim':
                if res.lower() != str(log_dict[key]).lower():
                    log_res[key] = False
                else:
                    log_res[key] = True
            elif res != str(log_dict[key]):
                log_res[key] = False
                if key == 'tuning':
                    # if you're tuning and weren't before
                    if log_dict['tuning']:
                        # then the old inflation is irrelevant
                        log_res['domain_inflation'] = False
                        log_res['hlg_inflation'] = False
                    # elif you're not tuning, were before, and inflations are the same
                    elif log_res['domain_inflation'] and log_res['hlg_inflation']:
                        log_res['tuning'] = True
                        log_dict['tuning'] = literal_eval(res)
                    # elif the domain inflation is the same and the hlg is not
 #                   else:
#                        log_res['hlg_inflation'] = False
            else:
                log_res[key] = True
                if key == 'tuning' and log_dict['tuning']:
                    if inflation_rnd1:
                        log_res['domain_inflation'] = True
                    else:
                        inflation_rnd2 = None
                        log_res['domain_inflation'] = False
                    if inflation_rnd2:
                        log_res['hlg_inflation'] = True
                    else:
                        log_res['hlg_inflation'] = False
            # if there's an inflation, tuning is complete
            if key == 'domain_inflation':
                if log_dict['tuning']:
                    inflation_rnd1 = literal_eval(res)
                else: # else the inflation is the inflation in the log_dict
                    inflation_rnd1 = log_dict['domain_inflation']
            elif key == 'hlg_inflation':
                # if you're tuning 
                if log_dict['tuning'] and log_res['hlg_inflation']:
                    inflation_rnd2 = literal_eval(res)
                else:
                    inflation_rnd2 = log_dict['hlg_inflation']
             

    init_discrep = []
    try:
        if not log_res['mtdb']:
            init_discrep.append('-d')
            log_res['n50'] = False
    except KeyError:
        init_discrep.append('-d')
    try:
        if not log_res['partition']:
            init_discrep.append('-nr/-np')
            log_res['null_samples'] = False
    except KeyError:
        init_discrep.append('-nr/-np')
    try:
        if not log_res['focal_genes']:
            init_discrep.append('-f')
            log_res['microsynt_tree'] = False
    except KeyError:
        init_discrep.append('-f')
    try:
        if not log_res['microsynt_constraint']:
            init_discrep.append('-c')
            log_res['microsynt_tree'] = False
    except KeyError:
        init_discrep.append('-c')
    try:
        if not log_res['n50']:
            init_discrep.append('--n50')
            log_res['plusminus'] = False
    except KeyError:
        init_discrep.append('--n50')
    try:
        if not log_res['dist_type']:
            init_discrep.append('-mmd')
            log_res['uniq_sp'] = False
    except KeyError:
        init_discrep.append('-mmd')
    try:
        if not log_res['uniq_sp']:
            init_discrep.append('-u')
            log_res['null_samples'] = False
    except KeyError:
        init_discrep.append('-u')
    try:
        if not log_res['plusminus']:
            init_discrep.append('-w')
            log_res['null_samples'] = False
    except KeyError:
        init_discrep.append('-w')
    try:
        if not log_res['null_samples']:
            init_discrep.append('-ns')
            log_res['hgp_percentile'] = False
    except KeyError:
        init_discrep.append('-ns')
    try:
        if not log_res['hgp_percentile']:
            init_discrep.append('-hp')
            log_res['hgx_percentile_I'] = False
    except KeyError:
        init_discrep.append('-hp')
    try:
        if not log_res['hgx_percentile_I']:
            init_discrep.append('-xp')
            log_res['hgx_percentile_II'] = False
    except KeyError:
        init_discrep.append('-xp')
    try:
        if not log_res['hgx_percentile_II']:
            init_discrep.append('-xp')
            log_res['topology_merge'] = False
    except KeyError:
        init_discrep.append('-xp')
    try:
        if not log_res['topology_merge']:
            log_res['min_topology_sim'] = False
            init_discrep.append('-tm')
    except KeyError:
        init_discrep.append('-tm')
    try:
        if not log_res['min_topology_sim']:
            init_discrep.append('-ts')
            log_res['hlg_sim'] = False
    except KeyError:
        init_discrep.append('-ts')
    try:
        if not log_res['aligner']:
            init_discrep.append('-a')
            log_res['gene_id'] = False
    except KeyError:
        log_res['aligner'] = False
        init_discrep.append('-a')
    try:
        if not log_res['gene_id']:
            init_discrep.append('-mg')
            log_res['hlg_sim'] = False
    except KeyError:
        init_discrep.append('-mg')
    try:
        if not log_res['hlg_sim']:
            init_discrep.append('-s')
            log_res['orig_hlg_id'] = False
    except KeyError:
        init_discrep.append('-s')
    try:
        if not log_res['orig_hlg_id']:
            init_discrep.append('-ml')
            log_res['domain_inflation'] = False
    except KeyError:
        init_discrep.append('-ml')
    try:
        if not log_res['domain_inflation']:
            init_discrep.append('-I2/-T')
            log_res['hlg_inflation'] = False
    except KeyError:
        init_discrep.append('-I1/-T')
        print('\nERROR: corrupted log.txt.' + \
            '\nIf not rectified, future runs will completely overwrite the current\n')
        sys.exit(149)
    try:
        if not log_res['hlg_inflation']:
            init_discrep.append('-I2/-T')
#            log_res['hlg_percentile'] = False
    except KeyError:
        init_discrep.append('-I2/-T')
        print('\nERROR: corrupted log.txt\nFuture runs may overwrite the current\n')
        sys.exit(149)
#    try:
 #       if not log_res['hlg_percentile']:
  #          init_discrep.append('-fp')
   # except KeyError:
    #    init_discrep.append('-fp')

    # stop gap for legacy data
    if 'hg_dir' not in log_res:
        log_res['hg_dir'] = True
    if 'hgx_dir' not in log_res:
        log_res['hgx_dir'] = True

    return log_res, inflation_rnd1, inflation_rnd2, init_discrep


def rm_old_data(
    log_res, out_dir, wrk_dir
    ):

    todel, tosave, save_dir = [], [], None
    if not log_res['dist_type'] or not log_res['uniq_sp']:
        if os.path.isfile(out_dir + 'working/omes2dist.pickle'):
            os.remove(out_dir + 'working/omes2dist.pickle')
    if not log_res['null_samples']:
        if os.path.isdir(wrk_dir + 'null/'):
            shutil.rmtree(wrk_dir + 'null/')
    if not log_res['hgp_percentile']:
        seed_file = out_dir + 'hgps.tsv.gz'
        seed_arr = wrk_dir + 'microsynt.npy'
        if os.path.isfile(seed_file):
            tosave.append(seed_file)
        if os.path.isfile(seed_arr):
            os.remove(seed_arr)
        clus_pickle = wrk_dir + 'hgx2loc.pickle'
        if os.path.isfile(clus_pickle):
            os.remove(clus_pickle)
    if not log_res['hgx_percentile_I']:
        skip_files = collect_files(wrk_dir + 'null/', '*')
        skip_files = [x for x in skip_files \
                      if os.path.basename(x).startswith('skip.')]
        for skip_file in skip_files:
            os.remove(skip_file)
        done_file = [x for x in collect_files(f'{wrk_dir}null/', 'txt') \
                     if os.path.basename(x).startswith('done')]
        if done_file:
            os.remove(done_file[0])
    if not log_res['hgx_percentile_II']:
        hlg_dir = wrk_dir + 'hlg/'
        if os.path.isdir(hlg_dir):
            todel.append(hlg_dir)
        groups = ['hgcs', 'gene2hgx']
        for group in groups:
             if os.path.isfile(f'{wrk_dir}{group}.pickle'):
                 os.remove(f'{wrk_dir}{group}.pickle')
    if not log_res['min_topology_sim']:
        groups = ['group.III']
        clan_data = ['hg.json.gz', 'hgx.json.gz', 'json.gz']
        for file_ in clan_data:
            if os.path.isfile(f'{wrk_dir}clan2loci.{file_}'):
                os.remove(f'{wrk_dir}clan2loci.{file_}')
        for group in groups:
             if os.path.isfile(f'{wrk_dir}{group}.pickle'):
                 os.remove(f'{wrk_dir}{group}.pickle')       
        todel.append(f'{wrk_dir}hlg/')
    if not log_res['aligner']:
        hgx_dir = f'{wrk_dir}hgx/'
        eprint('\tWARNING: alignment software changed, old alignments retained:', flush = True)
        eprint(f'\t\t{hgx_dir}', flush = True)
 #       shutil.rmtree(hgx_dir)
    if not log_res['orig_hlg_id']:
        hlg_dir = f'{wrk_dir}hlg/'
        adj_rows = collect_files(hlg_dir, 'tmp.w')
        adj_rows.extend(collect_files(hlg_dir, 'tmp.r'))
        adj_rows.extend(collect_files(hlg_dir, 'tmp'))
        row_file = wrk_dir + 'hlg/loci.adj'
        mci_file = wrk_dir + 'hlg/loci.mci'
        mcl_rows = wrk_dir + 'hlg/mcl_rows.tsv'
        for f in [row_file, mci_file, mcl_rows]:
            if os.path.isfile(f):
                todel.append(f)

    if not log_res['hlg_inflation']:
        hlgs_file = f'{wrk_dir}hlgs.pickle'
        if os.path.isfile(hlgs_file):
            os.remove(hlgs_file)
        mcl_res = f'{wrk_dir}hlg/loci.clus'
        tosave.extend([mcl_res, mcl_res + '.tmp'])
        hgx_files = ['gcl.pickle',
                     'mmi.pickle',
                     'mmp.pickle']
        for file_ in hgx_files:
            if os.path.isfile(f'{wrk_dir}{file_}'):
                os.remove(f'{wrk_dir}{file_}')

#    if not log_res['hlg_percentile']:
        hlg_output = out_dir + 'hlgs.tsv.gz'
        gcf_output = out_dir + 'gcfs.tsv.gz'
        kern_file = out_dir + 'hgxs.tsv.gz'
        ome_dir = out_dir + 'ome/'
        ann_dir = wrk_dir + 'ann/'
        tosave.extend([hlg_output, gcf_output])
        if os.path.isdir(ome_dir):
            count = 0
            save_dir = f'{out_dir}run{count}/'
            while os.path.isdir(save_dir):
                count += 1
                save_dir = f'{out_dir}run{count}/'
            os.mkdir(save_dir)
            os.mkdir(save_dir + 'working/')
            os.mkdir(save_dir + 'working/hlg/')
            os.mkdir(save_dir + 'working/hlg/lg')
            shutil.move(ome_dir, save_dir)
            if os.path.isfile(kern_file):
                shutil.copy(kern_file, save_dir + os.path.basename(kern_file))
            if os.path.isfile(hlg_output):
                shutil.move(hlg_output, save_dir + os.path.basename(hlg_output))
            if os.path.isfile(out_dir + 'hlgs.html'):
                shutil.move(out_dir + 'hlgs.html', save_dir + 'hlgs.html')
            if os.path.isfile(out_dir + 'hlg_metrics.html'):
                shutil.move(out_dir + 'hlg_metrics.html', 
                            save_dir + 'hlg_metrics.html')
            if os.path.isfile(gcf_output):
                shutil.move(gcf_output, save_dir + os.path.basename(clus_file))
            if os.path.isfile(out_dir + 'gcfs.html'):
                shutil.move(out_dir + 'gcfs.html', save_dir + 'gcfs.html')
            if os.path.isfile(out_dir + 'gcf_metrics.html'):
                shutil.move(out_dir + 'gcf_metrics.html', 
                            save_dir + 'gcf_metrics.html')

            shutil.move(out_dir + 'log.txt', save_dir + 'log.txt')
            for f in collect_files(f'{wrk_dir}hlg/', '*'):
                shutil.copy(f, f'{save_dir}working/hlg/{os.path.basename(f)}')
            for f in collect_files(f'{wrk_dir}hlg/lg/', '*'):
                shutil.copy(f, f'{save_dir}working/hlg/lg/{os.path.basename(f)}')

#        else:
 #           if os.path.isfile(clus_file):
  #              os.remove(clus_file)
    
#        if os.path.isdir(hmm_dir):
 #           shutil.rmtree(hmm_dir)
  #      if os.path.isdir(f'{out_dir}net/'):
   #         shutil.move(f'{out_dir}net/', f'{save_dir}net/')

    if not log_res['hg_dir']:
        if os.path.isdir(f'{wrk_dir}hg/'):
            shutil.rmtree(f'{wrk_dir}hg/')
    if not log_res['hgx_dir']:
        if os.path.isdir(f'{wrk_dir}hgx/'):
            shutil.rmtree(f'{wrk_dir}hgx/')

    if save_dir:
        for f in tosave:
            try:
                shutil.move(f, save_dir + os.path.basename(f))
            except FileNotFoundError:
                continue
    else:
        for f in tosave:
            try:
                os.remove(f)
            except FileNotFoundError:
                continue
    for f in todel:
        if os.path.isdir(f):
            shutil.rmtree(f)
        elif os.path.isfile(f):
            os.remove(f)

    if not log_res['hg_file']:
        shutil.rmtree(wrk_dir)
        os.mkdir(wrk_dir)
        for key in log_res:
            log_res[key] = False


def log_check(log_dict, log_path, out_dir, wrk_dir, flag = True):

    gcf_thresh = False
    if not os.path.isfile(log_path):
        log_res = {x: False for x in log_dict}
        try:
            rm_old_data(log_res, out_dir, wrk_dir)
        except KeyError:
            eprint('\nWARNING: missing log, rerunning without deleting', flush = True)
        init_log(log_path, log_dict)
    log_res, inflation_1, inflation_2, init_discrep = read_log(log_path, log_dict)
    if any(not log_res[x] for x in log_res):
        failed = set([x for x, v in log_res.items() if not v])
        for skip in ['patch_threshold', 'gcl_threshold', 'iipr_percent',
                     'id_percent', 'pos_percent', 'dist_threshold', 'ipr_path',
                     'pfam_path']:
            if skip in failed:
                failed.remove(skip)
        if failed:
            try:
                if not flag:
                    print(f'\n{init_discrep[0]} changed; removing data to checkpoint', flush = True)
                    rm_old_data(log_res, out_dir, wrk_dir)
                    init_log(log_path, log_dict)
                else:
                    eprint('\nERROR: -n not called and incompatible parameters: \
                            \n\t' + init_discrep[0], flush = True)
                    sys.exit(15)
            # irrelevant missing discrepancy
            except IndexError:
                init_log(log_path, log_dict)
        else:
            if os.path.isdir(out_dir + 'ome/'):
                tsvs = set(collect_files(out_dir + 'ome/', 'tsv', recursive = True))
                # NEED an intelligent resume
                ome_dirs = [f'{out_dir}ome/{x}/' for x in os.listdir(out_dir + 'ome/') \
                            if os.path.isdir(f'{out_dir}ome/{x}/')]
                if all(f'{x}hlg.tsv' in tsvs for x in ome_dirs) or os.path.isfile(out_dir + 'hlgs.tsv.gz'):
                    print('\nBypassing to filtration', flush = True)
                    gcf_thresh = True
                else:
                    init_log(log_path, log_dict)
            else:
                init_log(log_path, log_dict)
    elif os.path.isdir(out_dir + 'ome/'):
        tsvs = set(collect_files(out_dir + 'ome/', 'tsv', recursive = True))
        ome_dirs = [f'{out_dir}ome/{x}/' for x in os.listdir(out_dir + 'ome/') \
                    if os.path.isdir(f'{out_dir}ome/{x}/')]
        if all(f'{x}hlg.tsv' in tsvs for x in ome_dirs):
            print('\nBypassing to filtration', flush = True)
            gcf_thresh = True


    return log_res, inflation_1, inflation_2, gcf_thresh


def compileCDS(gff_list, ome):
    """ 
    Inputs the gff_list and organism ome code. Compiles CDS entries for loci parsing and outputs:
    cds_dict = {contig: {protein: [[CDSstart, CDSstop]] } }
    Sorts the cds_dict for each protein from smallest to largest
    """     
                
    cds_dict, fail = defaultdict(dict), False
    for entry in gff_list:
        if entry['type'] == 'CDS':
            prot_prep_i0 = entry['attributes'].index(';Alias=') # grab the
            # mycotools accession index
            try: # grab the end of the Alias by grabbing the index for a ';',
            # that is after the accession tag's occurrence.
            # then add that to the start of the accession tag and add one to
            # remove that semicolon. If there is a value error, the accession 
            # is at the end of the attributes (gff col 8 (?))
                prot_prep_i1 = entry['attributes'][prot_prep_i0+1:].index(';') + prot_prep_i0+1
            except ValueError: 
                prot_prep_i1 = len(entry['attributes'])
            prot = entry['attributes'][prot_prep_i0:prot_prep_i1].replace(';Alias=','')
            # obtain the protein accession
            if prot not in cds_dict[entry['seqid']] and prot: # add the protein
            # to the cds_dict entry
                cds_dict[entry['seqid']][prot] = [] 
            elif not prot: # if there isn't a valid accession it may mean the
            # mycotools curation did not work or the user did not curate
            # correctly 
                print(entry['attributes'], prot_prep_i0, prot_prep_i1)
                if not fail:
                    print('\tWARNING: ' + ome + ' has proteins in gff with no Alias', flush = True)
                fail = True
                continue    
            cds_dict[entry['seqid']][prot].extend([int(entry['start']),
                int(entry['end'])]) # add the coordinates
                        
    for contig in cds_dict:
        for prot in cds_dict[contig]:
            cds_dict[contig][prot].sort() # sort the coordinates of the proteins
            # lowest to highest
        cds_dict[contig] = list(sorted(cds_dict[contig].keys(), key = lambda k:
            cds_dict[contig][k][0])) # sort the proteins in the contigs lowest
            # to highest coordinates

    return dict(cds_dict)

def compileCDS2(gff_list, ome):
    """
    Inputs the gff_list and organism ome code. Compiles CDS entries for loci parsing and outputs:
    cds_dict = {contig: {protein: [[CDSstart, CDSstop]] } }
    Sorts the cds_dict for each protein from smallest to largest
    """

    cds_dict, cds_dict2, fail = defaultdict(dict), {}, False
    for entry in gff_list:
        if entry['type'] == 'CDS':
            prot_prep_i0 = entry['attributes'].index(';Alias=') # grab the
            # mycotools accession index
            try: # grab the end of the Alias by grabbing the index for a ';',
            # that is after the accession tag's occurrence.
            # then add that to the start of the accession tag and add one to
            # remove that semicolon. If there is a value error, the accession
            # is at the end of the attributes (gff col 8 (?))
                prot_prep_i1 = entry['attributes'][prot_prep_i0+1:].index(';') + prot_prep_i0+1
            except ValueError:
                prot_prep_i1 = len(entry['attributes'])
            prot = entry['attributes'][prot_prep_i0:prot_prep_i1].replace(';Alias=','')
            # obtain the protein accession
            if prot not in cds_dict[entry['seqid']] and prot: # add the protein
            # to the cds_dict entry
                cds_dict[entry['seqid']][prot] = []
                cds_dict2[prot] = []
            elif not prot: # if there isn't a valid accession it may mean the
            # mycotools curation did not work or the user did not curate
            # correctly
                print(entry['attributes'], prot_prep_i0, prot_prep_i1)
                if not fail:
                    print('\tWARNING: ' + ome + ' has proteins in gff with no Alias', flush = True)
                fail = True
                continue
            cds_dict[entry['seqid']][prot].extend([int(entry['start']),
                int(entry['end'])]) # add the coordinates
            cds_dict2[prot].append(entry)

    for contig in cds_dict:
        for prot in cds_dict[contig]:
            cds_dict[contig][prot].sort() # sort the coordinates of the proteins
            # lowest to highest
        cds_dict[contig] = list(sorted(cds_dict[contig].keys(), key = lambda k:
            cds_dict[contig][k][0])) # sort the proteins in the contigs lowest
            # to highest coordinates

    return dict(cds_dict), cds_dict2


def parseLoci(
    gff_path, ome_num, gene2hg, plusminus = 6
    ):
    """obtain a set of tuples of HG pairs {(OG0, OG1)...}"""

    gff_list = gff2list(gff_path) # open here to improve pickling
    cds_dict = compileCDS(gff_list, os.path.basename(gff_path).replace('.gff3',''))
    pairs = []
    for scaf in cds_dict: # for each contig
        for i, seq in enumerate(cds_dict[scaf]): # for each gene
            if i < plusminus: # does the locus start at the contig border?
                locus = cds_dict[scaf][:i+plusminus+1]
            else: # or I extend it to the +/- or opposite border
                locus = cds_dict[scaf][i-plusminus:i+plusminus+1]
            loc_og = []
            for gene in locus:
                try: # failure should be relatively rare, so faster than `if`
                    loc_og.append(gene2hg[gene])
                except KeyError: # it's either a singleton or not in an OG
                    pass
            pairs.extend([tuple(sorted(x)) for x in combinations(set(loc_og), 2)]) # don't deal with
            # same OGs - tandem duplications will just naturally cooccur more often

    out_pairs = set(pairs) # unique pairs of OGs
    return ome_num, out_pairs


def load_seedScores(file_):#, seed_thresh):

    out_hgs = []
    with gzip.open(file_, 'rt') as raw:
        for line in raw:
            if not line.startswith('#'):
                d = line.rstrip().split()
                out_hgs.append((int(d[0]), int(d[1])))
    
    return out_hgs


def compile_tree(i2ome, tree_path, root = []):
#    with open(tree_path, 'r') as raw:
 #       nwk = raw.read()

    phylo = load_tree(tree_path)
    omes = set(i2ome)
    tips = set(phylo.get_tip_names())
    missing = tips.difference(omes)
    if missing:
        eprint(f'\t\tWARNING: removing {len(missing)} ' \
              + 'missing tips from microsynteny tree',
               flush = True)
        todel = []
        for n in phylo.tips():
            if n.name in missing:
                todel.append(n)
        for n in todel:
            n.parent.remove(n)
            phylo.prune()
        
    if len(root) == 1:
        phylo = phylo.rooted_with_tip(root[0])
        phylo.write(tree_path, with_distances = True)
    elif len(root) > 1:
        nodes = {k: (v, len(v.get_tip_names())) \
                 for k, v in phylo.get_nodes_dict().items() \
                 if set(root).issubset(set(v.get_tip_names()))}
        mrca_tip_len = min([v[1] for v in list(nodes.values())])
        mrca_edge = [k for k, v in nodes.items() if v[1] == mrca_tip_len]
        try:
            phylo = phylo.rooted_at(mrca_edge[0])
        except:
            eprint(f'\nERROR: tree could not be rooted with supplied tips: ' \
                 + f'{root}', flush = True)
            sys.exit(449)
        phylo.write(tree_path, with_distances = True)

    phylo.reassign_names({v: str(i) for i, v in enumerate(i2ome)})
    return phylo


def write_hg_fa(hg_file, big_dict, genes):
    fa_str = dict2fa({k: big_dict[k] for k in genes})
    with open(f'{hg_file}.tmp', 'w') as out:
        out.write(fa_str)
    os.rename(f'{hg_file}.tmp', hg_file)


def output_hg_fas(db, genes, hg_file):
    fa_str = dict2fa(acc2fa(
                db, genes, error = False, spacer = '\t\t',
                coord_check = False
                ))
    with open(hg_file + '.tmp', 'w') as out:
        out.write(fa_str)
    os.rename(hg_file + '.tmp', hg_file)


def cp_files(f0, f1):
    shutil.copy(f0, f1)

def symlink_files(f0, f1):
    os.symlink(f0, f1)


def load_ome2fa(db, cpus = 1):
#    big_fa = {}
#    with mp.Pool(processes = cpus) as pool:
 #       fa_dicts = pool.map(fa2dict, 
  #                          tqdm((row['faa'] for row in db.values()), 
   #                         total = len(db)))
    #    pool.close()
     #   pool.join()
#    for res in fa_dicts:
 #       big_fa = {**big_fa, **res}
    fas = []
    for row in tqdm(db.values(), total = len(db)):
        fas.append(fa2dict(row['faa']))
    big_fa = {}
    for fa in fas:
        for nam, seq in fa.items():
            big_fa[nam] = seq

    return big_fa

def load_ome2fa_mp(db, cpus = 1):
    big_fa = {}
    with mp.Pool(processes = cpus) as pool:
        fa_dicts = pool.map(fa2dict, 
                            tqdm((row['faa'] for row in db.values()), 
                            total = len(db)))
        pool.close()
        pool.join()
    for res in fa_dicts:
        big_fa = {**big_fa, **res}

    return big_fa

def big_acc2fa(db, hg_dir, hgs, hg2gene, big_fa, cpus = 1):
    for hg in tqdm(hgs, total = len(hgs)):
        write_hg_fa(f'{hg_dir}{hg}.faa', big_fa, hg2gene[hg])



def big_acc2fa_mp(db, hg_dir, hgs, hg2gene, big_fa, cpus = 1):
    big_dict = mp.Manager().dict(big_fa)
    with mp.Pool(processes = cpus) as pool:
        pool.starmap(write_hg_fa, tqdm(((f'{hg_dir}{hg}.faa',
                                         big_dict, hg2gene[hg]) \
                                         for hg in hgs),
                                       total = len(hgs)))
        pool.close()
        pool.join()

    
def hg_fa_mngr(wrk_dir, hg_dir, hgs,
               db, hg2gene, cpus = 1, low_mem = False):
    new_hg_dir = wrk_dir + 'hg/'
    if not os.path.isdir(new_hg_dir):
        os.mkdir(new_hg_dir)
    # extract only used HGs
    hg_files = set([int(os.path.basename(x[:-4])) \
                    for x in collect_files(new_hg_dir, 'faa')])
    missing_hgs = sorted(set(hgs).difference(hg_files))

    if not missing_hgs:
        return new_hg_dir

    if not hg_dir:
        if low_mem: # this process is way too slow
            with mp.Pool(processes = cpus) as pool:
                pool.starmap(output_hg_fas,
                             tqdm(((db, hg2gene[hg], f'{new_hg_dir}{hg}.faa') \
                             for hg in missing_hgs), total = len(missing_hgs)))
                pool.close()
                pool.join()
        else:
            big_fa = load_ome2fa(db, cpus = cpus)
            big_acc2fa(db, new_hg_dir, missing_hgs, hg2gene, big_fa, cpus)
    else: # predetermined hg input
        hg_fa_cmds = []
        orthofinder = False
        if not os.path.isfile(f'{hg_dir}{hgs[0]}.faa'):
             digits = len(str(hgs[0]))
             zeros = 7 - digits
             if os.path.isfile(f'{hg_dir}OG{"0" * zeros}{hgs[0]}.fa'):
                 orthofinder = True
                 ext = '.fa'
             elif os.path.isfile(f'{hg_dir}{hgs[0]}.fa'):
                 ext = '.fa'
             elif os.path.isfile(f'{hg_dir}{hgs[0]}.fasta'):
                 ext = '.faa'
        else:
             raise FileNotFoundError('HG fastas missing')

        if orthofinder:
             for hg in missing_hgs:
                 digits = len(str(hg))
                 zeros = 7 - digits
                 hg_file = (f'{hg_dir}OG{"0" * zeros}{hg}.fa')
                 hg_fa_cmds.append((hg_file, f'{new_hg_dir}{hg}.faa'))
        else:
            hg_fa_cmds = [(f'{hg_dir}{hg}{ext}', f'{new_hg_dir}{hg}.faa') \
                          for hg in missing_hgs]
        copy_hgs = False
        try:
            os.symlink(hg_fa_cmds[0][0], hg_fa_cmds[0][1])
            del hg_fa_cmds[0]
        except:
            copy_hgs = True

        if copy_hgs:
            with mp.Pool(processes = cpus) as pool:
                pool.starmap(cp_files, hg_fa_cmds)
                pool.close()
                pool.join()
        else:
            with mp.Pool(processes = cpus) as pool:
                pool.starmap(symlink_files, hg_fa_cmds)
                pool.close()
                pool.join()
    return new_hg_dir


def sha_tune_file(tune_file):
    with open(tune_file, 'r') as raw:
        tune_data = [x.rstrip() for x in raw if not x.startswith('#')]
    tune = {}
    for cluster in tune_data:
        try:
            name, rawgenes, rawomes, rawfalse = cluster.split('\t')
        except ValueError:
            name, rawgenes, rawomes = cluster.split('\t')
            rawfalse = None
        if name in tune:
            eprint(f'\nERROR: duplicate entry for {name} in tune file', 
                   flush = True)
        if ',' in rawgenes:
            genes = [x.rstrip().lstrip() for x in rawgenes.split(',')]
        else:
            genes = [x.rstrip().lstrip() for x in rawgenes.split()]
        if ',' in rawomes:
            omes = [x.rstrip().lstrip() for x in rawomes.split(',')]
        else:
            omes = [x.rstrip().lstrip() for x in rawomes.split()]
        if rawfalse:
            if ',' in rawfalse:
                false_omes = [x.rstrip().lstrip() for x in rawfalse.split(',')]
            else:
                false_omes = [x.rstrip().lstrip() for x in rawfalse.split()]
        else:
            false_omes = []
        tune[name] = [tuple(sorted(set(genes))), tuple(sorted(set(omes))),
                      tuple(sorted(set(false_omes)))]
        
    tune_sha = hashlib.sha256(str(sorted(tune)).encode('utf-8')).hexdigest()
    return tune_sha


def init_run(db, out_dir, near_single_copy_genes, constraint_path,
             tree_path, hg_file, plusminus, seed_perc,# clus_perc,
             hgx_perc, aligner, id_perc, pos_perc, iipr_perc,
             patch_thresh, gcl_thresh, dist_thresh,
             samples, n50thresh, flag, min_gene_id, min_hlg_id, inflation_rnd1, 
             inflation_rnd2, simfun,
             tune_file, dist_type, uniq_sp, partition, min_topology_sim,
             topology_merge, hg_dir, hgx_dir, ipr_path, pfam_path):

    wrk_dir = out_dir + 'working/'
    if not os.path.isdir(wrk_dir):
        os.mkdir(wrk_dir)

    nul_dir = wrk_dir + 'null/'
    if not os.path.isdir(nul_dir):
        os.mkdir(nul_dir)

    log_path = out_dir + 'log.txt'
    if not tree_path:
        tree_path = out_dir + 'microsynt.newick'

    if tune_file:
        tune_sha = sha_tune_file(tune_file)
    else:
        tune_sha = None

    log_dict = {'mtdb': hashlib.sha256(str(sorted(db.keys())).encode('utf-8')).hexdigest(),
        'focal_genes': hashlib.sha256(','.join(sorted(
                                     near_single_copy_genes
                                     )).encode('utf-8')).hexdigest(),
        'microsynt_constraint': format_path(constraint_path),
        'microsynt_tree': tree_path,
        'hg_file': hg_file, 'plusminus': plusminus, 'dist_type': dist_type,
        'uniq_sp': bool(uniq_sp),
        'hgp_percentile': seed_perc, 'hgx_percentile': hgx_perc, 
        'aligner': aligner, 'gene_id': min_gene_id,
        'orig_hlg_id': None, 'hlg_id': min_hlg_id, 'hlg_sim': str(simfun).lower(),
#        'hlg_percentile': clus_perc, 
        'id_percent': id_perc, 'pos_percent': pos_perc, 'iipr_percent': iipr_perc,
        'patch_threshold': patch_thresh, 'dist_threshold': dist_thresh,
        'gcl_threshold': gcl_thresh, 'domain_inflation': inflation_rnd1,
        'hlg_inflation': inflation_rnd2,
        'tuning': tune_sha, 'partition': partition, 
        'topology_merge': topology_merge, 'min_topology_sim': min_topology_sim,
        'null_samples': samples, 'n50': n50thresh, 'hg_dir': hg_dir,
        'hgx_dir': hgx_dir, 'ipr_path': ipr_path, 'pfam_path': pfam_path}

    log_res, inflation_1, inflation_2, gcf_bypass = log_check(log_dict, log_path, 
                                                              out_dir, wrk_dir, flag)

    # symlink files on an individual basis so the new dir is writeable without
    # propagating new changes back to the reference dir
    hg_dir, hgx_dir = wrk_dir + 'hg/', wrk_dir + 'hgx/'

    existing_hg = set(os.path.basename(x) for x in collect_files(hg_dir, 'faa'))
    existing_hgx = set(os.path.basename(x) for x in collect_files(hgx_dir, 'out'))
    if log_dict['hg_dir']:
        if not os.path.isdir(hg_dir):
            os.mkdir(wrk_dir + 'hg/')
        prev_hg = set(os.path.basename(x) for x in collect_files(log_dict['hg_dir'], 'faa'))
        missing_hg = prev_hg.difference(existing_hg)
        for i in list(missing_hg):
            os.symlink(log_dict['hg_dir'] + i, wrk_dir + 'hg/' + i)
    if log_dict['hgx_dir']:
        if not os.path.isdir(hgx_dir):
            os.mkdir(wrk_dir + 'hgx/')
        prev_hgx = set(os.path.basename(x) for x in collect_files(log_dict['hgx_dir'], 'out'))
        missing_hgx = prev_hgx.difference(existing_hgx)
        for i in list(missing_hgx):
            os.symlink(log_dict['hgx_dir'] + i, wrk_dir + 'hgx/' + i)

    return wrk_dir, nul_dir, inflation_1, inflation_2, gcf_bypass



#! /usr/bin/env python3

import os, sys, argparse, multiprocessing as mp
from mycotools.lib.kontools import format_path, collect_files
from mycotools.lib.dbtools import mtdb, primaryDB

def CalcMeanMedian(genesPerClus):

    genesPerClus.sort()
    meanGpC = sum(genesPerClus)/len(genesPerClus)

    if len(genesPerClus) % 2: # it's odd
        medianGpC = int(genesPerClus[round((len(genesPerClus)/2))])
    else:
        medianGpC = sum(
            genesPerClus[int((len(genesPerClus)/2)-1):int((len(genesPerClus)/2)+1)]
            )/2


    return meanGpC, medianGpC

def GrabGenes_cloci(info_path):
    genes, gene2cluster, genesPerClus = [], {}, []
    with open(info_path, 'r') as raw:
        for line in raw:
            d = line.strip().split('\t')
            clus, tGenes = d[0], d[2].split(',')
            genes.extend(tGenes)
            genesPerClus.append(len(tGenes))
            for gene in tGenes:
                gene2cluster[gene] = clus

    if not genesPerClus: # something failed
        return 

    meanGpC, medianGpC = CalcMeanMedian(genesPerClus)

    return set(genes), gene2cluster, meanGpC, medianGpC

def GrabGenes_antismash(gbks):
    genes, gene2cluster, genesPerClus = [], {}, []
    for gbk in gbks:
        clus = os.path.basename(os.path.abspath(gbk))[:-4]
        with open(gbk, 'r') as raw:
            tGenes = []
            for line in raw:
                line = line.lstrip().rstrip()
                if line.startswith('/Alias'):
                    tGene = line[line.find('"'):]
                    gene = tGene.replace('"','')
                    tGenes.append(gene)
                    gene2cluster[gene] = clus
            genesPerClus.append(len(tGenes))
            genes.extend(tGenes)

    if not genesPerClus:
        return
    meanGpC, medianGpC = CalcMeanMedian(genesPerClus)

    return set(genes), gene2cluster, meanGpC, medianGpC

def GrabHits(clociGenes, antismashGenes, gene2cluster):
    intersection = clociGenes.intersection(antismashGenes)

    hits = []
    for gene in list(intersection):
        hits.append(gene2cluster[gene])

    return list(set(hits))

def ome_clus_mngr(ome, cloci_dir, as_dir):
    print(ome, flush = True)
    try:
        o2cGenes, o2cGene2Cluster, o2cMean, o2cMedian = GrabGenes_cloci(
            cloci_dir + 'ome/' + ome + '/info.out'
            )
    except IndexError: # failed
        return

    as_check = False
    if as_dir:
        if os.path.isdir(as_dir + ome):
            as_dir = as_dir + ome
            as_check = True

    asHits, o2cHits, asGene2Cluster, asMean, asMedian = \
        [], [], {}, None, None
    if as_check:
        gbks = collect_files(as_dir, 'gbk')
        gbks = [
            x for x in gbks if not \
            os.path.basename(os.path.abspath(x)).startswith(ome)
            ] # remove full gbk
        if gbks:
            asGenes, asGene2Cluster, asMean, asMedian = GrabGenes_antismash(gbks)
            o2cHits = GrabHits(o2cGenes, asGenes, o2cGene2Cluster)
            asHits = GrabHits(o2cGenes, asGenes, asGene2Cluster)

    stats = tuple([
        tuple(['o2c', tuple([
            len(set(o2cGene2Cluster.values())),
            o2cMean, o2cMedian, len(o2cHits)
            ])]),
        tuple(['as', tuple([
            len(set(asGene2Cluster.values())),
            asMean, asMedian, len(asHits)
            ])])
        ])


    return ome, stats

def write_stat_results(results, out_file):
    data = {
        k: v for k,v in sorted({
            x[0]: x[1] for x in results
            }.items(), key = lambda y: y[0]
            )
        } # sort stats by ome
    header = [
        'ome', 'clus', 'mean_genes_per', 
        'median_genes_per' #,0 'o2cRecoveredByAS', 'asClus', 
#        'asMeanGenesPerClus', 'asMedianGenesPerClus', 'asRecoveredByOG2clus'
        ]
    header_str = '#' + '\t'.join(header) + '\n'
    with open(out_file, 'w') as out:
        out.write(header_str)
        for ome, stats in data.items():
            if not ome: # it failed
                continue
            d = [
                ome, stats[0][1][0], stats[0][1][1], stats[0][1][2]
                ]
#            if stats[1][1][0]: # if there are antismash clusters
 #               d.extend([
  #                  stats[0][1][3]/stats[0][1][0],
   #                 stats[1][1][0], #stats[1][1][1], stats[1][1][2],
#                    stats[1][1][3]/stats[1][1][0]
    #            ])
    #        else:
      #          d.extend(['', '', '', '', ''])
            out.write('\t'.join([
                str(x) for x in d
                ]) + '\n')

def calc_cluster_stats(db, cloci_dir, cpus = 1):
    calc_stats_cmds = []
    for ome in db.set_index():
        if os.path.isdir(cloci_dir + 'ome/' + ome):
            calc_stats_cmds.append(
                [ome, cloci_dir, None]
                )

    with mp.Pool(processes = cpus) as pool:
        res = pool.starmap(ome_clus_mngr, calc_stats_cmds)

    write_stat_results(res, cloci_dir + 'stats.tsv')

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description = 'Summarize cloci output')
    parser.add_argument('-d', '--db', default = primaryDB())
    parser.add_argument('-c', '--cloci', required = True, help = 'cloci directory')
#    parser.add_argument('-a', '--antismash', help = 'antiSMASH directory')
    parser.add_argument('--cpu', default = 1, type = int)
    args = parser.parse_args()

    db = mtdb(format_path(args.db))
    o2c_dir = format_path(args.cloci)
 #   if args.antismash:
  #      as_dir = format_path(args.antismash)
   # else:
    #    as_dir = None

    calc_cluster_stats(db, o2c_dir, args.cpu)

    sys.exit(0)



#! /usr/bin/env python3

from collections import Counter, defaultdict
from itertools import combinations
import multiprocessing as mp, os, argparse, sys, re
import networkx as nx
from igraph import Graph
from pyvis.network import Network
from mycotools.lib.kontools import format_path, getColors, hex2rgb, write_json, read_json, eprint, mkOutput
from mycotools.lib.dbtools import mtdb
from mycotools.lib.biotools import gff2list
from mycotools.utils.extractHmmAcc import grabAccs
import matplotlib.pyplot as plt

# NEED to output json of the network modules
# NEED to organize output
# NEED to reference working dir pfams if it exists

def collectGffs(omes, top_dir):
    clusGFFs = [
        top_dir + ome + '/' + ome + '.gff3' for ome in omes \
        if os.path.isfile(top_dir + ome + '/' + ome + '.gff3')
        ]
    return clusGFFs

def mkPfam2name(pfam_file):
    # compile pfam 2 name
    with open(pfam_file, 'r') as raw:
        pfam_str = raw.read()
    pfams = grabAccs(pfam_str)
    pfam2name = {x[1].lstrip(): x[0].lstrip() for x in pfams}
    pfam2name = {k[:k.find('.')]: v for k,v in pfam2name.items()}
    return pfam2name


def extractLineages(db, rank):
    # acquire all lineages that belong to rank within input database
    omesXlineage = {}
    for ome, row in db.set_index().items():
        lineage = row['taxonomy'][rank]
        if lineage not in omesXlineage:
            omesXlineage[lineage] = []
        omesXlineage[lineage].append(ome)
    return omesXlineage

def parseClusGff(gff_list, ome):
    clus = {}
    for entry in gff_list:
        clusID = re.search(r'clusID=([^;]+)', entry['attributes'])[1]
        if clusID not in clus:
            clus[clusID] = []
        pfamSearch = re.search(r'PFAM=([^;]+)', entry['attributes'])
        if pfamSearch is not None:
            pfams = pfamSearch[1].replace('"','').split('|')
            clus[clusID].extend(pfams)

    clus = {
        k: sorted(list(set(v))) for k, v in sorted(clus.items(), key = lambda x: x[0])
        }

    counts, pfamCounts = Counter(), Counter()
    for clusID, pfams in clus.items():
        counts += Counter(combinations(pfams, 2))
        pfamCounts += Counter(pfams)

    return ome, tuple((k, v) for k, v in counts.items()), tuple((k, v) for k, v in pfamCounts.items())

#


def parseInfo(info_path, ome, gene2pfam):
    # acquire all orthogroup pairs from each cluster, return ready for counter hash

    counts, ogCounts, og2pfam = Counter(), Counter(), {}
    with open(info_path, 'r') as raw:
        for line in raw:
            todel = []
            if not line.startswith('#'):
                data = line.rstrip().split('\t')
                ogs, genes, pfams = [], data[2].split(','), []
                for i, og in enumerate(data[1].split(',')):
                    try:
                        ogs.append(int(og))
                        try:
                            pfams.append(gene2pfam[genes[i]])
                        except KeyError:
                            pfams.append([None])
                    except ValueError:
                        pass
#                try:
 #                   ogs = [int(x) for x in data[1].split(',')]
  #              except ValueError: # no og for a gene
  #                  todel.append(i)
#                try:
 #                   pfams = [x.split('|') for x in data[4].split(',')]
  #              except IndexError:
   #                 eprint(
    #                    '\t\tERROR: corrupted file/missing annotations -', 
     #                   ome, flush = True
      #                  )
       #             break
                if todel:
                    for i in reversed(todel):
                        del pfams[i]

                for i, og in enumerate(ogs):
                    if og not in og2pfam:
                        og2pfam[og] = []
                    try:
                        if pfams[i][0]:
                            og2pfam[og].extend(list(set(pfams[i])))
                    except IndexError:
                        pass
                sortOGs = sorted(ogs)
                ogCounts += Counter(set(sortOGs))
                pairs = combinations(sortOGs, 2)
                counts += Counter(pairs)
                

    return ome, tuple([tuple([k, v]) for k,v in counts.items()]), \
        tuple([tuple([k,v]) for k,v in ogCounts.items()]), \
        tuple([tuple([k,tuple(v)]) for k,v in og2pfam.items()])

def parsePfam(ome, pfam_res):
    gene2pfam = defaultdict(dict)
    pfam2def = {}
    with open(pfam_res, 'r') as raw:
        for line in raw:
            if line.startswith('#'):
                continue
            data = line.rstrip().split()
            g, p, q, b = data[0], data[3], data[2], float(data[5])
            # this is where a post-hoc threshold would be applied
            if q not in gene2pfam[g]: # in case there is a duplicate
            # for the same gene, add in the better score 
                gene2pfam[g][q] = b
            else:
                if b > gene2pfam[g][q]:
                    gene2pfam[g][q] = b
            pfam2def[p[:p.find('.')]] = q

#    for g,ps in gene2pfam.items():
 #       ps = [k for k,v in sorted(ps.items(), key = lambda x: x[1], reverse = True)][0]
        # grab the best bit for the gene

    return ome, tuple([tuple([k, tuple(list(v.keys()))]) for k, v in gene2pfam.items()])



def combineGffs(results, db, pfam2name, rank = 'species'):
    db = db.set_index()

    counts, pfamCounts, taxonLens = {}, {}, {}
    for ome, mCounts, mPfams in results:
        taxon = db[ome]['taxonomy'][rank]
        if taxon not in counts: 
            counts[taxon], taxonLens[taxon], pfamCounts[taxon] = Counter(), 0, Counter()
        taxonLens[taxon] += 1
        tCounts = {}
        for codes, val in mCounts:
            names = []
            for acc in codes:
                if acc[:acc.find('.')] in pfam2name:
                    names.append(pfam2name[acc[:acc.find('.')]])
                else:
                    names.append(acc)
            tCounts[tuple(sorted(names))] = val


        tPfams = {}
        for pfam, val in mPfams:
            abbr = pfam[:pfam.find('.')]
            if abbr in pfam2name:
                name = pfam2name[abbr]
            else:
                name = pfam
            tPfams[name] = val
    
        pfamCounts[taxon] += Counter(tPfams)
        counts[taxon] += Counter(tCounts)

    fCounts, fPfams = Counter(), Counter()
    for taxon in counts:
        fCounts += Counter({
            k: round(v/taxonLens[taxon]) for k, v in counts[taxon].items()
            })
        fPfams += Counter({
            k: round(v/taxonLens[taxon]) for k, v in pfamCounts[taxon].items()
            })

    return fCounts, fPfams

def combineCounts(countsRes, db, rank = 'species'):
    # combine counts from multiprocessing
    db = db.set_index()

    counts, ogCounts, og2pfam, taxonLens = {}, {}, {}, {}
    for ome, tCounts, tOGcounts, tOG2Pfam in countsRes:
        taxon = db[ome]['taxonomy'][rank]
        if taxon not in counts:
            counts[taxon], ogCounts[taxon] = Counter(), Counter()
            og2pfam[taxon], taxonLens[taxon] = {}, 0
        taxonLens[taxon] += 1
        counts[taxon] += Counter({x[0]: x[1] for x in tCounts})
        ogCounts[taxon] += Counter({x[0]: x[1] for x in tOGcounts})
        for og, pfams in tOG2Pfam:
            if og not in og2pfam[taxon]:
                og2pfam[taxon][og] = Counter()
            og2pfam[taxon][og] += Counter(pfams)

    fCounts, fOGcounts, fOG2Pfam = Counter(), Counter(), {}
    for taxon in counts:
        tCounts = {k: round(v/taxonLens[taxon]) for k, v in counts[taxon].items()}
        tOGcounts = {k: round(v/taxonLens[taxon]) for k, v in ogCounts[taxon].items()}
        tOG2Pfam = {}
        for og, pfams in og2pfam[taxon].items():
            tOG2Pfam[og] = {k: round(v/taxonLens[taxon]) for k, v in pfams.items()}

        fCounts += Counter(tCounts)
        fOGcounts += Counter(tOGcounts)
        for og, pfams in tOG2Pfam.items():
            if og not in fOG2Pfam:
                fOG2Pfam[og] = Counter()
            fOG2Pfam[og] += pfams

    for og in fOG2Pfam:
        fOG2Pfam[og] = {k: v for k,v in sorted(fOG2Pfam[og].items(), key = lambda x: x[1], reverse = True)}
            
    return fCounts, fOGcounts, fOG2Pfam

def writeNodeQuants(ogCounts, out_path):

    minC, maxC = min(ogCounts.values()), max(ogCounts.values())
    denom = maxC - minC
    ogCounts = [
        (k, v) for k,v in sorted(ogCounts.items(), key = lambda x: x[1],
        reverse = True)
        ]
    with open(out_path, 'w') as out:
        for k,v in ogCounts:
            out.write(str(k) + '\t' + str(((v-minC)/denom) + 0.01) + '\n')


def increaseFloat(val):
    # high fidelity float increase
    strVal = str(val)
    if strVal.endswith('9'):
        newVal = float(strVal + '1')
    else:
        factor = len(strVal[strVal.find('.')+1:])
        scale = 10**factor
        transVal = round(val*scale)
        newVal = float('0.' + str(transVal + 1))
    return newVal


def writeToAdj(oEdges, edges, out_path):
    minCount, maxCount = min(oEdges.values()), max(oEdges.values())
    denom = maxCount - minCount
    if denom == 0:
        eprint('\t\t\tWARNING: low connection weight throughout network', flush = True)
        denom = 1
    with open(out_path, 'w') as outF:
        for k, v in edges.items():
            value = ((v-minCount)/denom)
            out_data = [k[0], k[1], value]
            outF.write('\t'.join([str(x) for x in out_data]) + '\n')



def writeAdjMatr(oEdges, out_path, maximum = 500, duplicates = False):
    # write adjancency matrix of counts normalized scaled with factor

    if not duplicates:
        oEdges = {k: v for k,v in oEdges.items() if k[0] != k[1]}

    data = sorted(list(oEdges.values()))
    edges = oEdges
    writeToAdj(oEdges, edges, out_path + '.raw')

    percentile = 0.0
    while len(edges) > maximum: # while too many edges
        percentile = increaseFloat(percentile)
        minPerc = data[round((len(data)-1)*percentile)]
        edges = {k: 4*v for k, v in oEdges.items() if v > minPerc}

    print('\t\tEdge #|percentile:', len(edges), percentile, flush = True)
    writeToAdj(oEdges, edges, out_path)


def writeNetworkFile(
    adjMtx_path, quant_path, network_path, og2pfam_path,
    edgeConnect = False, centrality = False
    ):
    # open adjacency matrix and construct a network as an html

    quants = {}
    with open(quant_path, 'r') as raw:
        for line in raw:
            k, v = line.rstrip().split('\t')
            quants[k] = round(float(v)*4750)


    if os.path.isfile(og2pfam_path):
        og2pfamPrep = read_json(og2pfam_path)
        og2pfam = {}
        for og, pfams in og2pfamPrep.items():
            if pfams:
                v = list(pfams.keys())[0]
                og2pfam[og] = og + '|' + v[v.find('-')+1:]
            else:
                og2pfam[og] = og 
    else:
        og2pfam = {x: x for x in quants}

    if len(quants) < 2:
        eprint('\t\tERROR: less than 2 nodes in network', flush = True)
        return



    if os.path.isfile(adjMtx_path + '.raw'):
        print('\t\tQuantifying graph summary', flush = True)
        G = nx.read_weighted_edgelist(adjMtx_path + '.raw')

        if edgeConnect:
            print('\t\t\tQuantifying normalized graph connectivity', flush = True)
            print('\t\t\t\tNode:', flush = True)
            nodeConnectivity = nx.average_node_connectivity(G)
            print('\t\t\t\t\t' + str(nodeConnectivity), flush = True)
            print('\t\t\t\tEdge:', flush = True)
            edgeConnectivity = nx.edge_connectivity(G)
            print('\t\t\t\t\t' + str(edgeConnectivity), flush = True)
        elif centrality:
            print('\t\t\t\tQuantifying betweenness centrality', flush = True)
            cent_dict = nx.betweenness_centrality(G)
            central_path = adjMtx_path[:-4] + '.centrality'
            with open(central_path, 'w') as out:
                out.write('\n'.join([
                    str(og2pfam[k]) + '\t' + str(v) + '\t' + str(v/(quants[k]/4750))\
                    for k,v in sorted(
                        cent_dict.items(), key = lambda x: x[1], 
                        reverse = True
                        ) if v > 0 and k in quants # why isnt k always in quants
                    ]))
    net = nx.read_weighted_edgelist(adjMtx_path)

#    bb = nx.betweenness_centrality(net)
#    nx.set_node_attributes(net, bb, "betweenness")
    nx.set_node_attributes(net, quants, "size") # verify node sizes are applied right
    pos = nx.nx_pydot.pydot_layout(net)
#    pos = nx.get_node_attributes(net, 'pos')
    nodeSize = nx.get_node_attributes(net, 'size')
    nodeQuant = len(nodeSize)
    edgeLabels = nx.get_edge_attributes(net, 'weight')
    edgeQuant = len(edgeLabels)
    edgeLabels = {k: 4*v for k, v in edgeLabels.items()}
    
    print('\t\tNodes:', nodeQuant, flush = True)
    print('\t\tEdges:', len(edgeLabels), flush = True)

    print('\t\tCommunity detection', flush = True)
    # modularity detection via multilevel algorithm
#    h = Graph.from_networkx(net)
    h = Graph.TupleList(net.edges(), directed = False)
    com = Graph.community_multilevel(h, list(edgeLabels.values())) # verify edgelabels are extracted right
    membership = com.membership
    modulesPrep = {k: v for k, v in zip(h.vs["name"], membership)}
    modules = {}
    for og, clus in modulesPrep.items():
        if clus not in modules:
            modules[clus] = []
        modules[clus].append(og)
    modules = {k: v for k,v in sorted(modules.items(), key = lambda x: len(x[1]), reverse = True)}

    print('\t\t\tModules w/> 1 OG:', len([v for v in modules.values() if len(v) > 1]), flush = True)

    dimScale = 1+round(edgeQuant**0.65)
    if dimScale > 650:
        print('\t\tWARNING: extremely large network may not format well', flush = True)
        deScale = 650/dimScale
        dimScale = 650
        quants = {k: v*deScale for k,v in quants.items()}
        nx.set_node_attributes(net, quants, "size")
        edgeLabels = {k: v*deScale for k,v in edgeLabels.items()}


    fig = plt.figure(figsize=(dimScale,dimScale))
    ax = plt.subplot(111)

# just draw it outright to get all edges in there
    nx.draw(
        net, pos, node_size=list(nodeSize.values()), node_color='white' #, font_size=8,
#        with_labels = True, font_weight='bold'
        )

    colors = getColors(len([k for k, v in modules.items() if len(v) > 1]), ignore = ['#ffffff'])
    count, edgeCount, edgeColor = 0, -1, 'black'
    for module, nodes in modules.items():
        if len(nodes) > 1:
            try:
                color = colors[count]
                count += 1
                if edgeCount > -1:
                    edgeColor = colors[edgeCount]
                else:
                    edgeColor = 'black'
            except IndexError: # exceeds color
                count = 0
                color = colors[count]
                edgeCount += 1
                edgeColor = colors[edgeCount]
        else:
            color = '#ffffff'
            edgeColor = 'black'
        nx.draw_networkx_nodes(
            nx, pos, nodelist = tuple(nodes), node_color = color,
            node_size = tuple([quants[k] for k in nodes])
            #, font_size = 8, font_weight = 'bold', with_labels = True
            ).set_edgecolor(edgeColor)
        rgb = hex2rgb(color)
        if (rgb[0]*0.299 + rgb[1]*0.587 + rgb[2]*0.114) > 186: 
           fontColor = '#000000' 
        else:
           fontColor = '#aaaaaa'
        nx.draw_networkx_labels(
            nx, pos, labels = {k: og2pfam[k] for k in nodes}, font_size = 8, font_weight = 'bold', font_color = fontColor
            ) 

# edge labels
#    nx.draw_networkx_edge_labels(
 #       net, pos, edge_labels = edgeLabels, 
  #      font_size = 6
   #     )


    nx.draw_networkx_edges(
        net, pos, edgelist = edgeLabels.keys(), width = list(edgeLabels.values()),
        alpha = 0.4
        )
    try:
        plt.tight_layout()
    except ValueError: # too big
        ax.update_datalim(((0,0), (65535, 65535)))

    plt.show(block = True)
    plt.savefig(network_path, format="svg")
    plt.close()

# pyvis
#    nt = Network('500px', '500px')
 #   nt.from_nx(net)
  #  nt.show(network_path)

def main(
    db, rank, cloci_dir, gff_dir, connectivity = False, 
    centrality = False, maxEdges = 500, pfam_path = None, 
    pfam_dir = None, cpus = 1
    ):

    if cloci_dir:
        ome_dir, net_dir = cloci_dir + 'ome/', cloci_dir + 'net/'
        if not os.path.isdir(ome_dir):
            eprint('\nERROR: invalid cloci output', flush = True)
            sys.exit(2)
    if gff_dir:
        pfam2name = mkPfam2name(pfam_path)
        net_dir = mkOutput(gff_dir, 'clus2net')

    if not os.path.isdir(net_dir):
        os.mkdir(net_dir)
    omesXlineage = extractLineages(db, rank)

    omesXlineage = {
        k: v for k,v in sorted(omesXlineage.items(), key = lambda x: len(x[1]))
        }
    for lineage, omes in omesXlineage.items():
        print('\n' + lineage, flush = True)
        adjPath = net_dir + lineage + '.adj'
        quantPath = net_dir + lineage + '.tsv'
        og2pfamPath = net_dir + lineage + '.og2pfam.txt'
        if not os.path.isfile(adjPath): # need to log the factor then
            print('\tAdjacency matrix', flush = True)
            if cloci_dir:
                parsePfam_cmds = [
                    [ome, pfam_dir + ome + '.out'] for ome in omes
                    ]
                with mp.Pool(processes = cpus) as pool:
                    pfamRes = pool.starmap(parsePfam, parsePfam_cmds)
                ome2gene2pfam = {x[0]: dict(x[1]) for x in pfamRes}
                info_cmds = [
                    [ome_dir + ome + '/info.out', ome, ome2gene2pfam[ome]] \
                    for ome in omes \
                    if os.path.isfile(ome_dir + ome + '/info.out')
                    ]
                with mp.Pool(processes = cpus) as pool:
                    countsRes = pool.starmap(parseInfo, info_cmds)
                counts, ogCounts, og2pfam = combineCounts(
                    countsRes, db, rank = 'species'
                    )
                write_json(og2pfamPath, og2pfam)
                writeNodeQuants(ogCounts, quantPath)
            else:
                gffs = collectGffs(omes, gff_dir)
                with mp.Pool(processes = cpus) as pool:
                    results = pool.starmap(
                        parseClusGff, [
                            [gff2list(x), os.path.basename(x)[:-5]] \
                            for x in gffs
                            ]
                        )
                counts, pfamCounts = combineGffs(results, db, pfam2name, rank = 'species')
                writeNodeQuants(pfamCounts, quantPath)
            writeAdjMatr(counts, adjPath, maxEdges)
        
        netPath = adjPath[:-4] + '.svg'
        if not os.path.isfile(netPath):
            print('\tNetwork', flush = True)
            writeNetworkFile(adjPath, quantPath, netPath, og2pfamPath, connectivity, centrality)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description = 'Takes cloci output, creates networks of OG' + \
        ' relationships'
        )
    parser.add_argument(
        '-i', '--input', required = True,
        help = 'cloci results directory/antiSMASH results directory' + \
        " w/extracted .gff's"
        )
    parser.add_argument(
        '-d', '--database', required = True, 
        help = 'Must be same as cloci input'
        )
    parser.add_argument(
        '-r', '--rank', help = 'Taxon rank; DEFAULT: class',
        default = 'class'
        )
    parser.add_argument(
        '-m', '--maximum', help = 'Find upper percentile of edges to fit ' + \
        'maximum edges; DEFAULT: 500',
        default = 500, type = int
        )
    parser.add_argument(
       '-pd', '--pfam_dir', help = 'Directory of pfam tbl results labeled <OME.out>',
       required = True
       )
    parser.add_argument(
        '-p', '--pfam', help = 'Pfam-A.hmm for antiSMASH'
        )
    parser.add_argument(
        '--centrality', help = 'Calculate betweenness centrality',
        action = 'store_true'
        )
    parser.add_argument(
        '--connectivity', help = 'Quantify node and edge connectivity',
        action = 'store_true'
        )
    parser.add_argument(
        '-c', '--cpu', default = 1, type = int
        )
    args = parser.parse_args()

    db = mtdb(format_path(args.database))
    clus_dir = format_path(args.input)
    if not args.pfam:
        cloci_dir = clus_dir
        antismash_dir = None
    else:
        cloci_dir = None
        antismash_dir = clus_dir
    main(
        db, args.rank, cloci_dir, antismash_dir, connectivity = args.connectivity, 
        centrality = args.centrality,
        maxEdges = args.maximum, pfam_path = args.pfam, pfam_dir = format_path(args.pfam_dir), cpus = args.cpu
        )

